#! /usr/bin/env python

"""
2d fitting and creation of synthetic PSFs.
"""


__author__ = 'Carlos Alberto Gomez Gonzalez, V. Christiaens'
__all__ = ['create_synth_psf',
           'fit_2dgaussian',
           'fit_2dmoffat',
           'fit_2dairydisk',
           'fit_2d2gaussian']

import numpy as np
import pandas as pd
import photutils
from hciplot import plot_frames
from astropy.modeling import models, fitting
from astropy.stats import (gaussian_sigma_to_fwhm, gaussian_fwhm_to_sigma,
                           sigma_clipped_stats)
from .shapes import get_square, frame_center
from ..conf import check_array


def create_synth_psf(model='gauss', shape=(9, 9), amplitude=1, x_mean=None,
                     y_mean=None, fwhm=4, theta=0, gamma=None, alpha=1.5,
                     radius=None, msdi=False):
    """ Creates a synthetic 2d or 3d PSF with a 2d model: Airy disk, Gaussian or
    Moffat, depending on ``model``.

    Parameters
    ----------
    model : {'gauss', 'moff', 'airy'}, str optional
        Model to be used to create the synthetic PSF.
    shape : tuple of ints, optional
        Shape of the output 2d array.
    amplitude : float, optional
        Value of the amplitude of the 2d distribution.
    x_mean : float or None, optional
        Value of the centroid in X of the distributions: the mean of the
        Gaussian or the location of the maximum of the Moffat or Airy disk
        models. If None, the centroid is placed at the center of the array.
    y_mean : float or None, optional
        Value of the centroid in Y of the distributions: the mean of the
        Gaussian or the location of the maximum of the Moffat or Airy disk
        models. If None, the centroid is placed at the center of the array.
    fwhm : float, tuple of floats, list or np.ndarray, optional
        FWHM of the model in pixels. For the Gaussian case, it controls the
        standard deviation of the Gaussian. If a tuple is given, then the
        Gaussian will be elongated (fwhm in x, fwhm in y). For the Moffat, it is
        related to the gamma and alpha parameters. For the Airy disk, it is
        related to the radius (of the first zero) parameter. If ``msdi`` is True
        then ``fwhm`` must be a list of 1d np.ndarray (for example for
        SPHERE/IFS this sounds like a reasonable FWHM: np.linspace(4.5,6.7,39)).
    theta : float, optional
        Rotation angle in degrees of the Gaussian.
    gamma : float or None, optional
        Gamma parameter of core width of the Moffat model. If None, then it is
        calculated to correspond to the given ``fwhm``.
    alpha : float, optional
        Power index of the Moffat model.
    radius : float or None, optional
        The radius of the Airy disk (radius of the first zero). If None, then it
        is calculated to correspond to the given ``fwhm``.
    msdi : bool, optional
        Creates a 3d PSF, for emulating an IFS PSF.

    Returns
    -------
    im : numpy ndarray
        2d array with given ``shape`` and containing the synthetic PSF.

    Notes
    -----
    http://docs.astropy.org/en/stable/api/astropy.modeling.functional_models.Gaussian2D.html
    http://docs.astropy.org/en/stable/api/astropy.modeling.functional_models.Moffat2D.html
    http://docs.astropy.org/en/stable/api/astropy.modeling.functional_models.AiryDisk2D.html

    https://www.gnu.org/software/gnuastro/manual/html_node/PSF.html
    web.ipac.caltech.edu/staff/fmasci/home/astro_refs/PSFtheory.pdf
    web.ipac.caltech.edu/staff/fmasci/home/astro_refs/PSFsAndSampling.pdf
    """
    # 2d case
    if not msdi:
        sizex, sizey = shape
        if x_mean is None or y_mean is None:
            y_mean, x_mean = frame_center(np.zeros((sizey, sizex)))
        x = np.arange(sizex)
        y = np.arange(sizey)
        x, y = np.meshgrid(x, y)

        if model == 'gauss':
            if isinstance(fwhm, (tuple, list)):
                fwhm_x, fwhm_y = fwhm
            else:
                fwhm_y = fwhm
                fwhm_x = fwhm
            gauss = models.Gaussian2D(amplitude=amplitude, x_mean=x_mean,
                                      y_mean=y_mean,
                                      x_stddev=fwhm_x * gaussian_fwhm_to_sigma,
                                      y_stddev=fwhm_y * gaussian_fwhm_to_sigma,
                                      theta=np.deg2rad(theta))
            im = gauss(x, y)
        elif model == 'moff':
            if gamma is None and fwhm is not None:
                gamma = fwhm / (2. * np.sqrt(2 ** (1 / alpha) - 1))
            moffat = models.Moffat2D(amplitude=amplitude, x_0=x_mean,
                                     y_0=y_mean, gamma=gamma, alpha=alpha)
            im = moffat(x, y)
        elif model == 'airy':
            if radius is None and fwhm is not None:
                diam_1st_zero = (fwhm * 2.44) / 1.028
                radius = diam_1st_zero / 2.
            airy = models.AiryDisk2D(amplitude=amplitude, x_0=x_mean,
                                     y_0=y_mean, radius=radius)
            im = airy(x, y)
        return im
    # 3d case
    else:
        if not isinstance(fwhm, (list, np.ndarray)):
            raise ValueError('`Fwhm` must be a 1d vector')

        cube = []
        for fwhm_i in fwhm:
            cube.append(create_synth_psf(model, shape, amplitude, x_mean,
                                         y_mean, fwhm_i, theta, gamma, alpha,
                                         radius))
        cube = np.array(cube)
        return cube


def fit_2dgaussian(array, crop=False, cent=None, cropsize=15, fwhmx=4, fwhmy=4,
                   theta=0, threshold=False, sigfactor=6, full_output=True,
                   debug=True):
    """ Fitting a 2D Gaussian to the 2D distribution of the data.

    Parameters
    ----------
    array : numpy ndarray
        Input frame with a single PSF.
    crop : bool, optional
        If True an square sub image will be cropped.
    cent : tuple of int, optional
        X,Y integer position of source in the array for extracting the subimage.
        If None the center of the frame is used for cropping the subframe (the
        PSF is assumed to be ~ at the center of the frame).
    cropsize : int, optional
        Size of the subimage.
    fwhmx, fwhmy : float, optional
        Initial values for the standard deviation of the fitted Gaussian, in px.
    theta : float, optional
        Angle of inclination of the 2d Gaussian counting from the positive X
        axis.
    threshold : bool, optional
        If True the background pixels (estimated using sigma clipped statistics)
        will be replaced by small random Gaussian noise.
    sigfactor : int, optional
        The background pixels will be thresholded before fitting a 2d Gaussian
        to the data using sigma clipped statistics. All values smaller than
        (MEDIAN + sigfactor*STDDEV) will be replaced by small random Gaussian
        noise.
    full_output : bool, optional
        If False it returns just the centroid, if True also returns the
        FWHM in X and Y (in pixels), the amplitude and the rotation angle,
        and the uncertainties on each parameter.
    debug : bool, optional
        If True, the function prints out parameters of the fit and plots the
        data, model and residuals.

    Returns
    -------
    mean_y : float
        Source centroid y position on input array from fitting.
    mean_x : float
        Source centroid x position on input array from fitting.

    If ``full_output`` is True it returns a Pandas dataframe containing the
    following columns:
        'centroid_y': Y coordinate of the centroid.
        'centroid_x': X coordinate of the centroid.
        'fwhm_y': Float value. FHWM in X [px].
        'fwhm_x': Float value. FHWM in Y [px].
        'amplitude': Amplitude of the Gaussian.
        'theta': Float value. Rotation angle. 
        # and fit uncertainties on the above values: 
        'centroid_y_err'
        'centroid_x_err'
        'fwhm_y_err'
        'fwhm_x_err'
        'amplitude_err' 
        'theta_err'

    """
    check_array(array, dim=2, msg='array')

    if crop:
        if cent is None:
            ceny, cenx = frame_center(array)
        else:
            cenx, ceny = cent

        imside = array.shape[0]
        psf_subimage, suby, subx = get_square(array, min(cropsize, imside),
                                              ceny, cenx, position=True)
    else:
        psf_subimage = array.copy()

    if threshold:
        _, clipmed, clipstd = sigma_clipped_stats(psf_subimage, sigma=2)
        indi = np.where(psf_subimage <= clipmed + sigfactor * clipstd)
        subimnoise = np.random.randn(psf_subimage.shape[0],
                                     psf_subimage.shape[1]) * clipstd
        psf_subimage[indi] = subimnoise[indi]

    # Creating the 2D Gaussian model
    init_amplitude = np.ptp(psf_subimage)
    xcom, ycom = photutils.centroid_com(psf_subimage)
    gauss = models.Gaussian2D(amplitude=init_amplitude, theta=theta,
                              x_mean=xcom, y_mean=ycom,
                              x_stddev=fwhmx * gaussian_fwhm_to_sigma,
                              y_stddev=fwhmy * gaussian_fwhm_to_sigma)
    # Levenberg-Marquardt algorithm
    fitter = fitting.LevMarLSQFitter()
    y, x = np.indices(psf_subimage.shape)
    fit = fitter(gauss, x, y, psf_subimage)

    if crop:
        mean_y = fit.y_mean.value + suby
        mean_x = fit.x_mean.value + subx
    else:
        mean_y = fit.y_mean.value
        mean_x = fit.x_mean.value
    fwhm_y = fit.y_stddev.value*gaussian_sigma_to_fwhm
    fwhm_x = fit.x_stddev.value*gaussian_sigma_to_fwhm
    amplitude = fit.amplitude.value
    theta = np.rad2deg(fit.theta.value)
    
    # compute uncertainties
    if fitter.fit_info['param_cov'] is not None:
        perr = np.sqrt(np.diag(fitter.fit_info['param_cov']))
        amplitude_e, mean_x_e, mean_y_e, fwhm_x_e, fwhm_y_e, theta_e = perr
        fwhm_x_e /= gaussian_fwhm_to_sigma
        fwhm_y_e /= gaussian_fwhm_to_sigma
    else:
        amplitude_e, theta_e, mean_x_e = None, None, None
        mean_y_e, fwhm_x_e, fwhm_y_e = None, None, None

    if debug:
        if threshold:
            label = ('Subimage thresholded', 'Model', 'Residuals')
        else:
            label = ('Subimage', 'Model', 'Residuals')
        plot_frames((psf_subimage, fit(x, y), psf_subimage-fit(x, y)),
                    grid=True, grid_spacing=1, label=label)
        print('FWHM_y =', fwhm_y)
        print('FWHM_x =', fwhm_x, '\n')
        print('centroid y =', mean_y)
        print('centroid x =', mean_x)
        print('centroid y subim =', fit.y_mean.value)
        print('centroid x subim =', fit.x_mean.value, '\n')
        print('amplitude =', amplitude)
        print('theta =', theta)

    if full_output:
        return pd.DataFrame({'centroid_y': mean_y, 'centroid_x': mean_x,
                             'fwhm_y': fwhm_y, 'fwhm_x': fwhm_x,
                             'amplitude': amplitude, 'theta': theta,
                             'centroid_y_err': mean_y_e, 
                             'centroid_x_err': mean_x_e,
                             'fwhm_y_err': fwhm_y_e, 'fwhm_x_err': fwhm_x_e,
                             'amplitude_err': amplitude_e, 
                             'theta_err': theta_e}, index=[0],
                            dtype=np.float64)
    else:
        return mean_y, mean_x


def fit_2dmoffat(array, crop=False, cent=None, cropsize=15, fwhm=4,
                 threshold=False, sigfactor=6, full_output=True, debug=True):
    """ Fitting a 2D Moffat to the 2D distribution of the data.

    Parameters
    ----------
    array : numpy ndarray
        Input frame with a single PSF.
    crop : bool, optional
        If True an square sub image will be cropped.
    cent : tuple of int, optional
        X,Y integer position of source in the array for extracting the subimage.
        If None the center of the frame is used for cropping the subframe (the
        PSF is assumed to be ~ at the center of the frame).
    cropsize : int, optional
        Size of the subimage.
    fwhm : float, optional
        Initial values for the FWHM of the fitted 2d Moffat, in px.
    threshold : bool, optional
        If True the background pixels (estimated using sigma clipped statistics)
        will be replaced by small random Gaussian noise.
    sigfactor : int, optional
        The background pixels will be thresholded before fitting a 2d Moffat
        to the data using sigma clipped statistics. All values smaller than
        (MEDIAN + sigfactor*STDDEV) will be replaced by small random Gaussian
        noise.
    full_output : bool, optional
        If False it returns just the centroid, if True also returns the
        FWHM in X and Y (in pixels), the amplitude and the rotation angle.
    debug : bool, optional
        If True, the function prints out parameters of the fit and plots the
        data, model and residuals.

    Returns
    -------
    mean_y : float
        Source centroid y position on input array from fitting.
    mean_x : float
        Source centroid x position on input array from fitting.

    If ``full_output`` is True it returns a Pandas dataframe containing the
    following columns:
    'alpha': Float value. Alpha parameter.
    'amplitude' : Float value. Moffat Amplitude.
    'centroid_x' : Float value. X coordinate of the centroid.
    'centroid_y' : Float value. Y coordinate of the centroid.
    'fwhm' : Float value. FHWM [px].
    'gamma' : Float value. Gamma parameter.

    """
    check_array(array, dim=2, msg='array')

    if crop:
        if cent is None:
            ceny, cenx = frame_center(array)
        else:
            cenx, ceny = cent

        imside = array.shape[0]
        psf_subimage, suby, subx = get_square(array, min(cropsize, imside),
                                              ceny, cenx, position=True)
    else:
        psf_subimage = array.copy()

    if threshold:
        _, clipmed, clipstd = sigma_clipped_stats(psf_subimage, sigma=2)
        indi = np.where(psf_subimage <= clipmed + sigfactor * clipstd)
        subimnoise = np.random.randn(psf_subimage.shape[0],
                                     psf_subimage.shape[1]) * clipstd
        psf_subimage[indi] = subimnoise[indi]

    # Creating the 2D Moffat model
    init_amplitude = np.ptp(psf_subimage)
    xcom, ycom = photutils.centroid_com(psf_subimage)
    moffat = models.Moffat2D(amplitude=init_amplitude, x_0=xcom, y_0=ycom,
                             gamma=fwhm / 2., alpha=1)
    # Levenberg-Marquardt algorithm
    fitter = fitting.LevMarLSQFitter()
    y, x = np.indices(psf_subimage.shape)
    fit = fitter(moffat, x, y, psf_subimage)

    if crop:
        mean_y = fit.y_0.value + suby
        mean_x = fit.x_0.value + subx
    else:
        mean_y = fit.y_0.value
        mean_x = fit.x_0.value

    fwhm = fit.fwhm
    amplitude = fit.amplitude.value
    alpha = fit.alpha.value
    gamma = fit.gamma.value

    if debug:
        if threshold:
            label = ('Subimage thresholded', 'Model', 'Residuals')
        else:
            label = ('Subimage', 'Model', 'Residuals')
        plot_frames((psf_subimage, fit(x, y), psf_subimage - fit(x, y)),
                    grid=True, grid_spacing=1, label=label)
        print('FWHM =', fwhm)
        print('centroid y =', mean_y)
        print('centroid x =', mean_x)
        print('centroid y subim =', fit.y_0.value)
        print('centroid x subim =', fit.x_0.value, '\n')
        print('amplitude =', amplitude)
        print('alpha =', alpha)
        print('gamma =', gamma)

    # compute uncertainties
    if fitter.fit_info['param_cov'] is not None:
        perr = np.sqrt(np.diag(fitter.fit_info['param_cov']))
        amplitude_err, mean_x_err, mean_y_err, gamma_err, alpha_err = perr
        fwhm_err = 2*gamma_err
    else:
        amplitude_err, mean_x_err, mean_y_err = None, None, None
        gamma_err, alpha_err, fwhm_err  = None, None, None
    
    if full_output:
        return pd.DataFrame({'centroid_y': mean_y, 'centroid_x': mean_x,
                             'fwhm': fwhm, 'alpha': alpha, 'gamma': gamma,
                             'amplitude': amplitude, 'centroid_y_err': mean_y_err, 
                             'centroid_x_err': mean_x_err,
                             'fwhm_err': fwhm_err, 'alpha_err': alpha_err, 
                             'gamma_err': gamma_err, 
                             'amplitude_err': amplitude_err}, index=[0],
                            dtype=np.float64)
    else:
        return mean_y, mean_x


def fit_2dairydisk(array, crop=False, cent=None, cropsize=15, fwhm=4,
                   threshold=False, sigfactor=6, full_output=True, debug=True):
    """ Fitting a 2D Airy to the 2D distribution of the data.

    Parameters
    ----------
    array : numpy ndarray
        Input frame with a single PSF.
    crop : bool, optional
        If True an square sub image will be cropped.
    cent : tuple of int, optional
        X,Y integer position of source in the array for extracting the subimage.
        If None the center of the frame is used for cropping the subframe (the
        PSF is assumed to be ~ at the center of the frame).
    cropsize : int, optional
        Size of the subimage.
    fwhm : float, optional
        Initial values for the FWHM of the fitted 2d Airy disk, in px.
    threshold : bool, optional
        If True the background pixels (estimated using sigma clipped statistics)
        will be replaced by small random Gaussian noise.
    sigfactor : int, optional
        The background pixels will be thresholded before fitting a 2d Moffat
        to the data using sigma clipped statistics. All values smaller than
        (MEDIAN + sigfactor*STDDEV) will be replaced by small random Gaussian
        noise.
    full_output : bool, optional
        If False it returns just the centroid, if True also returns the
        FWHM in X and Y (in pixels), the amplitude and the rotation angle.
    debug : bool, optional
        If True, the function prints out parameters of the fit and plots the
        data, model and residuals.

    Returns
    -------
    mean_y : float
        Source centroid y position on input array from fitting.
    mean_x : float
        Source centroid x position on input array from fitting.

    If ``full_output`` is True it returns a Pandas dataframe containing the
    following columns:
    'amplitude' : Float value. Moffat Amplitude.
    'centroid_x' : Float value. X coordinate of the centroid.
    'centroid_y' : Float value. Y coordinate of the centroid.
    'fwhm' : Float value. FHWM [px].

    """
    check_array(array, dim=2, msg='array')

    if crop:
        if cent is None:
            ceny, cenx = frame_center(array)
        else:
            cenx, ceny = cent

        imside = array.shape[0]
        psf_subimage, suby, subx = get_square(array, min(cropsize, imside),
                                              ceny, cenx, position=True)
    else:
        psf_subimage = array.copy()

    if threshold:
        _, clipmed, clipstd = sigma_clipped_stats(psf_subimage, sigma=2)
        indi = np.where(psf_subimage <= clipmed + sigfactor * clipstd)
        subimnoise = np.random.randn(psf_subimage.shape[0],
                                     psf_subimage.shape[1]) * clipstd
        psf_subimage[indi] = subimnoise[indi]

    # Creating the 2d Airy disk model
    init_amplitude = np.ptp(psf_subimage)
    xcom, ycom = photutils.centroid_com(psf_subimage)
    diam_1st_zero = (fwhm * 2.44) / 1.028
    airy = models.AiryDisk2D(amplitude=init_amplitude, x_0=xcom, y_0=ycom,
                             radius=diam_1st_zero/2.)
    # Levenberg-Marquardt algorithm
    fitter = fitting.LevMarLSQFitter()
    y, x = np.indices(psf_subimage.shape)
    fit = fitter(airy, x, y, psf_subimage)

    if crop:
        mean_y = fit.y_0.value + suby
        mean_x = fit.x_0.value + subx
    else:
        mean_y = fit.y_0.value
        mean_x = fit.x_0.value

    amplitude = fit.amplitude.value
    radius = fit.radius.value
    fwhm = ((radius * 1.028) / 2.44) * 2

    # compute uncertainties
    if fitter.fit_info['param_cov'] is not None:
        perr = np.sqrt(np.diag(fitter.fit_info['param_cov']))
        amplitude_err, mean_x_err, mean_y_err, radius_err = perr
        fwhm_err = ((radius_err * 1.028) / 2.44) * 2
    else:
        amplitude_err, mean_x_err, mean_y_err = None, None, None
        radius_err, fwhm_err = None, None

    if debug:
        if threshold:
            label = ('Subimage thresholded', 'Model', 'Residuals')
        else:
            label = ('Subimage', 'Model', 'Residuals')
        plot_frames((psf_subimage, fit(x, y), psf_subimage - fit(x, y)),
                    grid=True, grid_spacing=1, label=label)
        print('FWHM =', fwhm)
        print('centroid y =', mean_y)
        print('centroid x =', mean_x)
        print('centroid y subim =', fit.y_0.value)
        print('centroid x subim =', fit.x_0.value, '\n')
        print('amplitude =', amplitude)
        print('radius =', radius)

    if full_output:
        return pd.DataFrame({'centroid_y': mean_y, 'centroid_x': mean_x,
                             'fwhm': fwhm, 'radius': radius,
                             'amplitude': amplitude, 
                             'centroid_y_err': mean_y_err, 
                             'centroid_x_err': mean_x_err, 
                             'fwhm_err': fwhm_err, 'radius_err': radius_err,
                             'amplitude_err': amplitude_err}, index=[0])
    else:
        return mean_y, mean_x
        
        
def fit_2d2gaussian(array, crop=False, cent=None, cropsize=15, fwhm_neg=4, 
                    fwhm_pos=4, theta_neg=0, theta_pos=0, neg_amp=1, 
                    fix_neg=True, threshold=False, sigfactor=2, 
                    full_output=False, debug=True):
    """ Fitting a 2D superimposed double Gaussian (negative and positive) to 
    the 2D distribution of the data (reproduce e.g. the effect of a coronagraph)

    Parameters
    ----------
    array : numpy ndarray
        Input frame with a single PSF.
    crop : bool, optional
        If True an square sub image will be cropped.
    cent : tuple of float, optional
        X,Y position of the source in the array for extracting the 
        subimage. If None the center of the frame is used for cropping the 
        subframe. If fix_neg is set to True, this will also be used as the 
        fixed position of the negative gaussian.
    cropsize : int, optional
        Size of the subimage.
    fwhm_neg, fwhm_pos : float or tuple of floats, optional
        Initial values for the FWHM of the fitted negative and positive 
        Gaussians, in px. If a tuple, should be the FWHM value along x and y.
    theta_neg, theta_pos: float, optional
        Angle of inclination of the 2d Gaussian counting from the positive X
        axis (only matters if a tuple was provided for fwhm_neg or fwhm_pos).
    neg_amp: float, optional
        First guess on the amplitude of the negative gaussian, relative to the
        amplitude of the positive gaussian (i.e. 1 means the negative gaussian
        has the same amplitude as the positive gaussian)
    fix_neg: bool, optional
        Whether to fix the position and FWHM of the negative gaussian for a 
        fit with less free parameters. In that case, the center of the negative
        gaussian is assumed to be cent
    threshold : bool, optional
        If True the background pixels (estimated using sigma clipped statistics)
        will be replaced by small random Gaussian noise.
    sigfactor : int, optional
        The background pixels will be thresholded before fitting a 2d Gaussian
        to the data using sigma clipped statistics. All values smaller than
        (MEDIAN + sigfactor*STDDEV) will be replaced by small random Gaussian
        noise.
    full_output : bool, optional
        If False it returns just the centroid, if True also returns the
        FWHM in X and Y (in pixels), the amplitude and the rotation angle,
        and the uncertainties on each parameter.
    debug : bool, optional
        If True, the function prints out parameters of the fit and plots the
        data, model and residuals.

    Returns
    -------
    mean_y : float
        Source centroid y position on input array from fitting.
    mean_x : float
        Source centroid x position on input array from fitting.

    If ``full_output`` is True it returns a Pandas dataframe containing the
    following columns:
    - for the positive gaussian:
    'amplitude' : Float value. Amplitude of the Gaussian.
    'centroid_x' : Float value. X coordinate of the centroid.
    'centroid_y' : Float value. Y coordinate of the centroid.
    'fwhm_x' : Float value. FHWM in X [px].
    'fwhm_y' : Float value. FHWM in Y [px].
    'theta' : Float value. Rotation angle of x axis
    - for the negative gaussian:
    'amplitude_neg' : Float value. Amplitude of the Gaussian.
    'centroid_x_neg' : Float value. X coordinate of the centroid.
    'centroid_y_neg' : Float value. Y coordinate of the centroid.
    'fwhm_x_neg' : Float value. FHWM in X [px].
    'fwhm_y_neg' : Float value. FHWM in Y [px].
    'theta_neg' : Float value. Rotation angle of x axis
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array')

    if cent is None:
        ceny, cenx = frame_center(array)
    else:
        cenx, ceny = cent
        
    if crop:
        x_sub_px = cenx%1
        y_sub_px = ceny%1

        imside = array.shape[0]
        psf_subimage, suby, subx = get_square(array, min(cropsize, imside),
                                              int(ceny), int(cenx), 
                                              position=True)
        ceny, cenx = frame_center(psf_subimage)
        ceny+=y_sub_px
        cenx+=x_sub_px              
    else:
        psf_subimage = array.copy()

    if threshold:
        _, clipmed, clipstd = sigma_clipped_stats(psf_subimage, sigma=2)
        indi = np.where(psf_subimage <= clipmed + sigfactor * clipstd)
        subimnoise = np.random.randn(psf_subimage.shape[0],
                                     psf_subimage.shape[1]) * clipstd
        psf_subimage[indi] = subimnoise[indi]

    if isinstance(fwhm_neg, tuple):
        fwhm_neg_x, fwhm_neg_y = fwhm_neg
    else:
        fwhm_neg_x = fwhm_neg
        fwhm_neg_y = fwhm_neg
        
    if isinstance(fwhm_pos, tuple):
        fwhm_pos_x, fwhm_pos_y = fwhm_pos
    else:
        fwhm_pos_x = fwhm_pos
        fwhm_pos_y = fwhm_pos

    # Creating the 2D Gaussian model
    init_amplitude = np.ptp(psf_subimage)
    #xcom, ycom = photutils.centroid_com(psf_subimage)
    ycom, xcom = frame_center(psf_subimage)
    fix_dico_pos = {'theta':True}
    bounds_dico_pos = {'amplitude':[0.8*init_amplitude,1.2*init_amplitude],
                       'x_mean':[xcom-3,xcom+3],
                       'y_mean':[ycom-3,ycom+3],
                       'x_stddev': [0.5*fwhm_pos_x*gaussian_fwhm_to_sigma,
                                    2*fwhm_pos_x*gaussian_fwhm_to_sigma],
                       'y_stddev': [0.5*fwhm_pos_y*gaussian_fwhm_to_sigma,
                                    2*fwhm_pos_y*gaussian_fwhm_to_sigma]}
                                    
    gauss_pos = models.Gaussian2D(amplitude=init_amplitude, x_mean=xcom, 
                                  y_mean=ycom, 
                                  x_stddev=fwhm_pos_x*gaussian_fwhm_to_sigma, 
                                  y_stddev=fwhm_pos_y*gaussian_fwhm_to_sigma, 
                                  theta=np.deg2rad(theta_pos)%(np.pi), fixed=fix_dico_pos, 
                                  bounds=bounds_dico_pos)
    if fix_neg:                          
        fix_dico_neg = {'x_mean':True,'y_mean':True,
                        'x_stddev':True,
                        'y_stddev':True,
                        'theta':True}
        bounds_dico_neg = {'amplitude':[neg_amp*0.5*init_amplitude,
                                        neg_amp*2*init_amplitude]}
    else:
        fix_dico_neg = {}#{'theta':True}
        bounds_dico_neg = {'amplitude':[neg_amp*0.5*init_amplitude,
                                        neg_amp*2*init_amplitude],
                           'x_mean':[xcom-3,xcom+3],
                           'y_mean':[ycom-3,ycom+3],
                           'x_stddev': [0.5*fwhm_neg_x*gaussian_fwhm_to_sigma,
                                        2*fwhm_neg_x*gaussian_fwhm_to_sigma],
                           'y_stddev': [0.5*fwhm_neg_y*gaussian_fwhm_to_sigma,
                                        2*fwhm_neg_y*gaussian_fwhm_to_sigma],
                           'theta': [0,np.pi]}
    
    gauss_neg = models.Gaussian2D(amplitude=init_amplitude*neg_amp, x_mean=cenx, 
                                  y_mean=ceny, 
                                  x_stddev=fwhm_neg_x*gaussian_fwhm_to_sigma, 
                                  y_stddev=fwhm_neg_y*gaussian_fwhm_to_sigma, 
                                  theta=np.deg2rad(theta_neg)%(np.pi), fixed=fix_dico_neg, 
                                  bounds=bounds_dico_neg)

    double_gauss = gauss_pos-gauss_neg

    fitter = fitting.LevMarLSQFitter() #SLSQPLSQFitter() #LevMarLSQFitter()                  
    y, x = np.indices(psf_subimage.shape)                          
    fit = fitter(double_gauss, x, y, psf_subimage, maxiter=100000, acc=1e-08)
    
    # positive gaussian
    if crop:
        mean_y = fit.y_mean_0.value + suby
        mean_x = fit.x_mean_0.value + subx
    else:
        mean_y = fit.y_mean_0.value
        mean_x = fit.x_mean_0.value
    fwhm_y = fit.y_stddev_0.value*gaussian_sigma_to_fwhm
    fwhm_x = fit.x_stddev_0.value*gaussian_sigma_to_fwhm
    amplitude = fit.amplitude_0.value
    theta = np.rad2deg(fit.theta_0.value)

    # negative gaussian    
    if crop:
        mean_y_neg = fit.y_mean_1.value + suby
        mean_x_neg = fit.x_mean_1.value + subx
    else:
        mean_y_neg = fit.y_mean_1.value
        mean_x_neg = fit.x_mean_1.value
    fwhm_y_neg = fit.y_stddev_1.value*gaussian_sigma_to_fwhm
    fwhm_x_neg = fit.x_stddev_1.value*gaussian_sigma_to_fwhm
    amplitude_neg = fit.amplitude_1.value
    theta_neg = np.rad2deg(fit.theta_1.value)  
    
    if debug:
        if threshold:
            label = ('Subimage thresholded', 'Model', 'Residuals')
        else:
            label = ('Subimage', 'Model', 'Residuals')
        plot_frames((psf_subimage, fit(x, y), psf_subimage-fit(x, y)),
                     grid=True, grid_spacing=1, label=label)
        print('FWHM_y =', fwhm_y)
        print('FWHM_x =', fwhm_x, '\n')
        print('centroid y =', mean_y)
        print('centroid x =', mean_x)
        print('centroid y subim =', fit.y_mean_0.value)
        print('centroid x subim =', fit.x_mean_0.value, '\n')
        print('amplitude =', amplitude)
        print('theta =', theta)
        print('FWHM_y (neg) =', fwhm_y_neg)
        print('FWHM_x (neg) =', fwhm_x_neg, '\n')
        print('centroid y (neg) =', mean_y_neg)
        print('centroid x (neg) =', mean_x_neg)
        print('centroid y subim (neg) =', fit.y_mean_1.value)
        print('centroid x subim (neg) =', fit.x_mean_1.value, '\n')
        print('amplitude (neg) =', amplitude_neg)
        print('theta (neg) =', theta_neg)
        
    if full_output:
        return pd.DataFrame({'centroid_y': mean_y, 'centroid_x': mean_x,
                             'fwhm_y': fwhm_y, 'fwhm_x': fwhm_x,
                             'amplitude': amplitude, 'theta': theta,
                             'centroid_y_neg': mean_y_neg, 
                             'centroid_x_neg': mean_x_neg,
                             'fwhm_y_neg': fwhm_y_neg, 
                             'fwhm_x_neg': fwhm_x_neg,
                             'amplitude_neg': amplitude_neg, 
                             'theta_neg': theta_neg}, index=[0],
                            dtype=np.float64)
    else:
        return mean_y, mean_x


#!/usr/bin/env python

import os
import re
from setuptools import setup
try:  # pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:  # pip <= 9.0.3
    from pip.req import parse_requirements
from setuptools.command.install import install
from setuptools.command.develop import develop


# Hackishly override of the install method
class InstallReqs(install):
    def run(self):
        print(" ********************** ")
        print(" *** Installing VIP *** ")
        print(" ********************** ")
        os.system('pip install -r requirements.txt')
        install.run(self)


class InstallDevReqs(develop):
    def run(self):
        print(" **************************** ")
        print(" *** Installing VIP (dev) *** ")
        print(" **************************** ")
        os.system('pip install -r requirements-dev.txt')
        develop.run(self)


def resource(*args):
    return os.path.join(os.path.abspath(os.path.join(__file__, os.pardir)),
                        *args)


# parse_requirements() returns generator of pip.req.InstallRequirement objects
reqs = parse_requirements(resource('requirements.txt'), session=False)
try:
    reqs = [str(ir.req) for ir in reqs]
except:
    reqs = [str(ir.requirement) for ir in reqs]
reqs_dev = parse_requirements(resource('requirements-dev.txt'), session=False)
try:
    reqs_dev = [str(ir.req) for ir in reqs_dev]
except:
    reqs_dev = [str(ir.requirement) for ir in reqs_dev]    

with open(resource('README.rst')) as readme_file:
    README = readme_file.read()

with open(resource('vip_hci', '__init__.py')) as version_file:
    version_file = version_file.read()
    VERSION = re.search(r"""^__version__ = ['"]([^'"]*)['"]""",
                        version_file, re.M)
    VERSION = VERSION.group(1)


PACKAGES = ['vip_hci',
            'vip_hci.andromeda',
            'vip_hci.conf',
            'vip_hci.exlib',
            'vip_hci.fits',
            'vip_hci.frdiff',
            'vip_hci.leastsq',
            'vip_hci.llsg',
            'vip_hci.medsub',
            'vip_hci.metrics',
            'vip_hci.negfc',
            'vip_hci.nmf',
            'vip_hci.pca',
            'vip_hci.preproc',
            'vip_hci.specfit',
            'vip_hci.stats',
            'vip_hci.var']

setup(
    name='vip_hci',
    version=VERSION,
    description='Package for astronomical high-contrast image processing.',
    long_description=README,
    license='MIT',
    author='Carlos Alberto Gomez Gonzalez',
    author_email='carlosgg33@gmail.com',
    url='https://github.com/vortex-exoplanet/VIP',
    cmdclass={'install': InstallReqs,
              'develop': InstallDevReqs},
    packages=PACKAGES,
    install_requires=reqs,
    extras_require={"dev": reqs_dev},
    zip_safe=False,
    classifiers=['Intended Audience :: Science/Research',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: POSIX :: Linux',
                 'Natural Language :: English',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Topic :: Scientific/Engineering :: Astronomy'
                 ]
)



#! /usr/bin/env python

"""
Module with frame/cube filtering functionalities.
"""



__author__ = 'Carlos Alberto Gomez Gonzalez, Valentin Christiaens'
__all__ = ['frame_filter_highpass',
           'frame_filter_lowpass',
           'frame_deconvolution',
           'cube_filter_highpass',
           'cube_filter_lowpass',
           'cube_filter_iuwt']

import warnings
try:
    import cv2
    no_opencv = False
except ImportError:
    msg = "Opencv python bindings are missing."
    warnings.warn(msg, ImportWarning)
    no_opencv = True
import numpy as np
from scipy.ndimage import median_filter
from skimage.restoration import richardson_lucy
from astropy.convolution import (convolve_fft, convolve, Gaussian2DKernel, 
                                 Moffat2DKernel)
from astropy.convolution import interpolate_replace_nans as interp_nan
from astropy.stats import gaussian_fwhm_to_sigma
from ..exlib import iuwt
from ..conf import Progressbar


def cube_filter_iuwt(cube, coeff=5, rel_coeff=1, full_output=False):
    """
    Isotropic Undecimated Wavelet Transform filtering.

    Parameters
    ----------
    cube : numpy ndarray
        Input cube.
    coeff : int, optional
        Number of wavelet scales to be used in the decomposition.
    rel_coeff : int, optional
        Number of relevant coefficients. In other words how many wavelet scales
        will represent in a better way our data. One or two scales are enough
        for filtering our images.
    full_output : bool, optional
        If True, an additional cube with the multiscale decomposition of each
        frame will be returned.

    Returns
    -------
    cubeout : numpy ndarray
        Output cube with the filtered frames.

    If full_output is True the filtered cube is returned together with the a
    4d cube containing the multiscale decomposition of each frame.

    """
    cubeout = np.zeros_like(cube)
    cube_coeff = np.zeros((cube.shape[0], coeff, cube.shape[1], cube.shape[2]))
    n_frames = cube.shape[0]

    print('Decomposing frames with the Isotropic Undecimated Wavelet Transform')
    for i in Progressbar(range(n_frames)):
        res = iuwt.iuwt_decomposition(cube[i], coeff, store_smoothed=False)
        cube_coeff[i] = res
        for j in range(rel_coeff):
            cubeout[i] += cube_coeff[i][j]

    if full_output:
        return cubeout, cube_coeff
    else:
        return cubeout


def cube_filter_highpass(array, mode='laplacian', verbose=True, **kwargs):
    """
    Apply ``frame_filter_highpass`` to the frames of a 3d or 4d cube.

    Parameters
    ----------
    array : numpy ndarray
        Input cube, 3d or 4d.
    mode : str, optional
        ``mode`` parameter to the ``frame_filter_highpass`` function. Defaults
        to a Laplacian high-pass filter.
    verbose : bool, optional
        If ``True`` timing and progress bar are shown.
    **kwargs : dict
        Passed through to the ``frame_filter_highpass`` function.

    Returns
    -------
    filtered : numpy ndarray
        High-pass filtered cube.

    """
    array_out = np.empty_like(array)

    if array.ndim == 3:
        for i in Progressbar(range(array.shape[0]), verbose=verbose):
            array_out[i] = frame_filter_highpass(array[i], mode=mode, **kwargs)
    elif array.ndim == 4:
        for i in Progressbar(range(array.shape[1]), verbose=verbose):
            for lam in range(array.shape[0]):
                array_out[lam][i] = frame_filter_highpass(array[lam][i],
                                                          mode=mode, **kwargs)
    else:
        raise TypeError('Input array is not a 3d or 4d cube')

    return array_out


def fft(array):
    """
    Perform the 2d discrete Fourier transform, using numpy's fft2 function.

    This produces a new representation of
    the image in which each pixel represents a spatial frequency and
    orientation, rather than an xy coordinate. When Fourier-transformed images
    are plotted graphically, the low frequencies are found at the centre; this
    is not what fft2 actually produces, so we need to also apply numpy's
    fftshift (centering low frequencies).

    """
    fft_array = np.fft.fftshift(np.fft.fft2(array))
    return fft_array


def ifft(array):
    """
    Get the inverse Fourier transform on the image.

    This produces an array of complex numbers whose real values correspond to
    the image in the original space (decentering).

    Notes
    -----
    A real function corresponds to a symmetric function in fourier space. As
    long as the operations we apply in the fourier space do not break this
    symmetry, the data returned by ``ifft`` should not containy any imaginary
    part.

    """
    new_array = np.fft.ifft2(np.fft.ifftshift(array)).real
    return new_array


def frame_filter_highpass(array, mode, median_size=5, kernel_size=5,
                          fwhm_size=5, btw_cutoff=0.2, btw_order=2,
                          hann_cutoff=5, psf=None, gauss_mode='conv', 
                          mask=None):
    """
    High-pass filtering of input frame depending on parameter ``mode``.

    The filtered image properties will depend on the ``mode`` and the relevant
    parameters.

    Parameters
    ----------
    array : numpy ndarray
        Input array, 2d frame.
    mode : str
        Type of High-pass filtering.

        ``laplacian``
            applies a Laplacian fiter with kernel size defined by
            ``kernel_size`` using the Opencv library.
        ``laplacian-conv``
            applies a Laplacian high-pass filter by defining a kernel (with
            ``kernel_size``) and using the ``convolve_fft`` Astropy function.
        ``median-subt``
            subtracts a median low-pass filtered version of the image.
        ``gauss-subt``
            subtracts a Gaussian low-pass filtered version of the image.
        ``fourier-butter``
            applies a high-pass 2D Butterworth filter in Fourier domain.
        ``hann``
            uses a Hann window.

    median_size : int, optional
        Size of the median box for the ``median-subt`` filter.
    kernel_size : int, optional
        Size of the Laplacian kernel used in ``laplacian`` mode. It must be an
        positive odd integer value.
    fwhm_size : float, optional
        Size of the Gaussian kernel used in ``gaus-subt`` mode.
    btw_cutoff : float, optional
        Frequency cutoff for low-pass 2d Butterworth filter used in
        ``fourier-butter`` mode.
    btw_order : int, optional
        Order of low-pass 2d Butterworth filter used in ``fourier-butter`` mode.
    hann_cutoff : float
        Frequency cutoff for the ``hann`` mode.
    psf: numpy ndarray, optional
        Input normalised and centered psf, 2d frame. Should be provided if 
        mode is set to 'psf'.
    gauss_mode : {'conv', 'convfft'}, str optional
        'conv' uses the multidimensional gaussian filter from scipy.ndimage and
        'convfft' uses the fft convolution with a 2d Gaussian kernel.
    mask: numpy ndarray, optional
        Binary mask indicating where the low-pass filtered image should be 
        interpolated with astropy.convolution. This otion can be useful if the 
        low-pass filtered image is aimed to capture low-spatial frequency sky 
        signal, while avoiding a stellar halo (set to one in the binary mask).
        Note: only works with Gaussian kernel or PSF convolution.

    Returns
    -------
    filtered : numpy ndarray
        High-pass filtered image.

    """
    def butter2d_lp(size, cutoff, n=3):
        """
        Create low-pass 2D Butterworth filter.

        Function from PsychoPy library, credits to Jonathan Peirce, 2010

        Parameters
        ----------
        size : tuple
            size of the filter
        cutoff : float
            relative cutoff frequency of the filter (0 - 1.0)
        n : int, optional
            order of the filter, the higher n is the sharper
            the transition is.

        Returns
        -------
        numpy ndarray
          filter kernel in 2D centered

        """
        if not 0 < cutoff <= 1:
            raise ValueError('Cutoff frequency must be between 0 and 1.')

        if not isinstance(n, int):
            raise ValueError('n must be an integer >= 1.')

        rows, cols = size
        x = np.linspace(-0.5, 0.5, cols) * cols
        y = np.linspace(-0.5, 0.5, rows) * rows

        # An array with every pixel = radius relative to center
        radius = np.sqrt((x**2)[np.newaxis] + (y**2)[:, np.newaxis])
        # The filter
        f = 1 / (1 + (radius / cutoff)**(2*n))
        return f

    def round_away(x):
        """
        Round to the *nearest* integer, half-away-from-zero.

        Parameters
        ----------
        x : array-like

        Returns
        -------
        r_rounded : array-like (float)

        Notes
        -----
        IDL ``ROUND`` rounds to the *nearest* integer (commercial rounding),
        unlike numpy's round/rint, which round to the nearest *even*
        value (half-to-even, financial rounding) as defined in IEEE-754
        standard.

        """
        return np.trunc(x + np.copysign(0.5, x))

    # --------------------------------------------------------------------------

    if array.ndim != 2:
        raise TypeError("Input array is not a frame or 2d array.")

    if mask is not None and (mode!='psf-subt' and mode!='gauss-subt'):
        msg="Masking option only available for gauss-subt and psf-subt modes"
        raise TypeError(msg)

    if mode == 'laplacian':
        # Applying a Laplacian high-pass kernel
        if kernel_size % 2 == 0 or kernel_size < 0:
            raise ValueError("Kernel size must be an odd and positive value.")
        if not no_opencv:
            msg = "Opencv bindings are missing. Trying a convolution with a "
            msg += "Laplacian kernel instead."

        filtered = cv2.Laplacian(-array, cv2.CV_32F, ksize=kernel_size)

    elif mode == 'laplacian-conv':
        # Applying a Laplacian high-pass kernel defining a kernel and using
        # the convolve_fft Astropy function
        kernel3 = np.array([[-1, -1, -1],
                            [-1, 8, -1],
                            [-1, -1, -1]])
        kernel5 = np.array([[-4, -1, 0, -1, -4],
                            [-1, 2, 3, 2, -1],
                            [0, 3, 4, 3, 0],
                            [-1, 2, 3, 2, -1],
                            [-4, -1, 0, -1, -4]])
        kernel7 = np.array([[-10, -5, -2, -1, -2, -5, -10],
                            [-5, 0, 3, 4, 3, 0, -5],
                            [-2, 3, 6, 7, 6, 3, -2],
                            [-1, 4, 7, 8, 7, 4, -1],
                            [-2, 3, 6, 7, 6, 3, -2],
                            [-5, 0, 3, 4, 3, 0, -5],
                            [-10, -5, -2, -1, -2, -5, -10]])
        if kernel_size == 3:
            kernel = kernel3
        elif kernel_size == 5:
            kernel = kernel5
        elif kernel_size == 7:
            kernel = kernel7
        else:
            raise ValueError('Kernel size must be either 3, 5 or 7.')
        filtered = convolve_fft(array, kernel, normalize_kernel=False,
                                nan_treatment='fill')

    elif mode == 'median-subt':
        # Subtracting the low_pass filtered (median) image from the image itself
        medianed = frame_filter_lowpass(array, 'median',
                                        median_size=median_size)
        filtered = array - medianed

    elif mode == 'gauss-subt':
        # Subtracting the low_pass filtered (median) image from the image itself
        gaussed = frame_filter_lowpass(array, 'gauss', fwhm_size=fwhm_size,
                                       gauss_mode=gauss_mode, mask=mask)
        filtered = array - gaussed

    elif mode == 'psf-subt':
        if psf is None:
            raise TypeError("psf should be provided for psf-subt mode")
        # Subtracting the low_pass filtered (median) image from the image itself
        psfed = frame_filter_lowpass(array, 'psf', psf=psf, mask=mask)
        filtered = array - psfed

    elif mode == 'fourier-butter':
        # Designs an n-th order high-pass 2D Butterworth filter
        filt = 1 - butter2d_lp(array.shape, cutoff=btw_cutoff, n=btw_order)
        array_fft = fft(array)
        fft_new = array_fft * filt
        filtered = ifft(fft_new)
        
    elif mode == 'hann':
        # TODO: this code could be shortened using np.convolve
        # see http://scipy-cookbook.readthedocs.io/items/SignalSmooth.html

        # create a Hanning profile window cut at the chosen frequency:
        npix = array.shape[0]
        cutoff = npix/2 * hann_cutoff
        cutoff_inside = round_away(np.minimum(cutoff, (npix/2 - 1))).astype(int)
        winsize = 2*cutoff_inside + 1
        win1d = np.hanning(winsize)
        win = 1 - np.outer(win1d, win1d)

        array_fft = fft(array)
        # remove high spatial frequency along the Hann profile:
        array_fft[npix//2 - cutoff_inside: npix//2 + cutoff_inside + 1,
                  npix//2 - cutoff_inside: npix//2 + cutoff_inside + 1] *= win

        filtered = ifft(array_fft)

    else:
        raise TypeError('Mode not recognized.')

    return filtered


def frame_filter_lowpass(array, mode='gauss', median_size=5, fwhm_size=5,
                         gauss_mode='conv', kernel_sz=None, psf=None, mask=None, 
                         iterate=True, **kwargs):
    """
    Low-pass filtering of input frame depending on parameter ``mode``.

    Parameters
    ----------
    array : numpy ndarray
        Input array, 2d frame.
    mode : {'median', 'gauss', 'moff', 'psf'}, str optional
        Type of low-pass filtering.
    median_size : int, optional
        Size of the median box for filtering the low-pass median filter.
    fwhm_size : float, optional
        Size of the Gaussian kernel for the low-pass Gaussian filter.
    gauss_mode : {'conv', 'convfft'}, str optional
        'conv' uses the multidimensional gaussian filter from scipy.ndimage and
        'convfft' uses the fft convolution with a 2d Gaussian kernel.
    kernel_sz: int or None, optional
        Size of the kernel in pixels for 2D Gaussian and Moffat convolutions.
        If None, astropy.convolution will automatically consider 8*radius 
        kernel sizes.
    psf: numpy ndarray, optional
        Input normalised and centered psf, 2d frame. Should be provided if 
        mode is set to 'psf'.
    mask: numpy ndarray, optional
        Binary mask indicating where the low-pass filtered image should be 
        interpolated with astropy.convolution. This option can be useful if the 
        low-pass filtered image is aimed to capture low-spatial frequency sky 
        signal, while avoiding a stellar halo (set to one in the binary mask).
        Note: only works with Gaussian kernel or PSF convolution.
    iterate: bool, opt
        If the first convolution leaves nans, whether to continue replacing 
        nans by interpolation until they are all replaced.
    **kwargs : dict
        Passed through to the astropy.convolution.convolve or convolve_fft
        function.

    Returns
    -------
    filtered : numpy ndarray
        Low-pass filtered image.

    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array.')
    if not isinstance(median_size, int):
        raise ValueError('`Median_size` must be integer')
        
    if mask is not None:
        if mode== 'median':
            msg="Masking not available for median filter"
        if mask.shape != array.shape:
            msg = "Mask dimensions should be the same as array"
            raise TypeError(msg)

    if mode == 'median':
        # creating the low_pass filtered (median) image
        filtered = median_filter(array, median_size, mode='nearest')
    elif mode == 'gauss':
        # 2d Gaussian filter
        sigma = fwhm_size * gaussian_fwhm_to_sigma
        if gauss_mode == 'conv':
            filtered = convolve(array, Gaussian2DKernel(x_stddev=sigma, 
                                                        x_size=kernel_sz,
                                                        y_size=kernel_sz),
                                mask=mask, **kwargs)
            if iterate and np.sum(np.isnan(filtered))>0:
                filtered = interp_nan(filtered, 
                                      Gaussian2DKernel(x_stddev=sigma, 
                                                       x_size=kernel_sz,
                                                       y_size=kernel_sz),
                                      convolve=convolve)
        elif gauss_mode == 'convfft':
            # FFT Convolution with a 2d gaussian kernel created with Astropy.
            filtered = convolve_fft(array, Gaussian2DKernel(x_stddev=sigma, 
                                                            x_size=kernel_sz,
                                                            y_size=kernel_sz),
                                    mask=mask, **kwargs)
            if iterate and np.sum(np.isnan(filtered))>0:
                filtered = interp_nan(filtered, 
                                      Gaussian2DKernel(x_stddev=sigma, 
                                                       x_size=kernel_sz,
                                                       y_size=kernel_sz),
                                      convolve=convolve_fft, **kwargs)
        else:
            raise TypeError('2d Gaussian filter mode not recognized')
    elif mode == 'psf':
        if psf is None:
            raise TypeError('psf should be provided for convolution')
        elif psf.ndim != 2:
            raise TypeError('Input psf is not a frame or 2d array.')
        if psf.shape[-1] > array.shape[-1]:
            raise TypeError('Input psf is larger than input array. Crop.')
        # psf convolution
        filtered = convolve_fft(array, psf, mask=mask, **kwargs)
        if iterate and np.sum(np.isnan(filtered))>0:
            filtered = interp_nan(filtered, psf, convolve=convolve_fft, 
                                  **kwargs)
    else:
        raise TypeError('Low-pass filter mode not recognized')
        
    return filtered


def cube_filter_lowpass(array, mode='gauss', median_size=5, fwhm_size=5,
                        gauss_mode='conv', kernel_sz=None, verbose=True, 
                        psf=None, mask=None, iterate=True, **kwargs):
    """
    Apply ``frame_filter_lowpass`` to the frames of a 3d or 4d cube.

    Parameters
    ----------
    array : numpy ndarray
        Input cube, 3d or 4d.
    mode : str, optional
        See the documentation of the ``frame_filter_lowpass`` function.
    median_size : int, optional
        See the documentation of the ``frame_filter_lowpass`` function.
    fwhm_size : float, optional
        See the documentation of the ``frame_filter_lowpass`` function.
    gauss_mode : str, optional
        See the documentation of the ``frame_filter_lowpass`` function.
    kernel_sz: int, optional
        See the documentation of the ``frame_filter_lowpass`` function.
    verbose : bool, optional
        If True timing and progress bar are shown.
    psf: numpy ndarray, optional
        Input normalised and centered psf, 2d frame. Should be provided if 
        mode is set to 'psf'.
    mask: numpy ndarray, optional
        Binary mask indicating where the low-pass filtered image should be 
        interpolated with astropy.convolution. This otion can be useful if the 
        low-pass filtered image is aimed to capture low-spatial frequency sky 
        signal, while avoiding a stellar halo (set to one in the binary mask).
        Note: only works with Gaussian kernel or PSF convolution.
    iterate: bool, opt
        If the first convolution leaves nans, whether to continue replacing 
        nans by interpolation until they are all replaced.
    **kwargs : dict
        Passed through to the astropy.convolution.convolve or convolve_fft
        function.
        
    Returns
    -------
    filtered : numpy ndarray
        Low-pass filtered cube.

    """
    array_out = np.empty_like(array)

    if array.ndim == 3:
        for i in Progressbar(range(array.shape[0]), verbose=verbose):
            array_out[i] = frame_filter_lowpass(array[i], mode, median_size,
                                                fwhm_size, gauss_mode, 
                                                kernel_sz, psf, mask, iterate, 
                                                **kwargs)
    elif array.ndim == 4:
        for i in Progressbar(range(array.shape[1]), verbose=verbose):
            for lam in range(array.shape[0]):
                array_out[lam][i] = frame_filter_lowpass(array[lam][i], mode,
                                                         median_size, fwhm_size,
                                                         gauss_mode, kernel_sz, 
                                                         psf, mask, iterate, 
                                                         **kwargs)
    else:
        raise TypeError('Input array is not a 3d or 4d cube')

    return array_out


def frame_deconvolution(array, psf, n_it=30):
    """
    Iterative image deconvolution following the scikit-image implementation 
    of the Richardson-Lucy algorithm.
    
    Considering an image that has been convolved by the point spread function
    of an instrument, the algorithm will sharpen the blurred 
    image through a user-defined number of iterations, which changes the
    regularisation.

    Reference: William Hadley Richardson, “Bayesian-Based Iterative Method of 
    Image Restoration”, J. Opt. Soc. Am. A 27, 1593-1607 (1972), 
    DOI:10.1364/JOSA.62.000055

    See also description at:
    https://en.wikipedia.org/wiki/Richardson%E2%80%93Lucy_deconvolution
    

    Parameters
    ----------
    array : numpy ndarray
        Input image, 2d frame.
    psf : numpy ndarray
        Input psf, 2d frame.    
    n_it : int, optional
        Number of iterations.

    Returns
    -------
    deconv : numpy ndarray
        Deconvolved image.

    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array.')
    if psf.ndim != 2:
        raise TypeError('Input psf is not a frame or 2d array.')

    max_I = np.amax(array)
    min_I = np.amin(array)
    drange = max_I-min_I

    deconv = richardson_lucy((array-min_I)/drange, psf, iterations=n_it)
    deconv*=drange
    deconv+=min_I

    return deconv


#! /usr/bin/env python

"""
Probability of a point source to be a background star.
"""

__author__ = 'V. Christiaens'
__all__ = ['bkg_star_proba']

from scipy.special import factorial    
import numpy as np


def bkg_star_proba(n_dens, sep, n_bkg=1, verbose=True, full_output=False):
    """ Given an input density of background star brighter than a certain 
    magnitude (obtained e.g. from the Besancon model or TRILEGAL), and the 
    separation of n_bkg point source, estimate the probability of having n_bkg 
    or more background stars in a disk with radius equal to the largest 
    separation.
    The probability is estimated using a spatial Poisson point process.

    Parameters
    ----------
    n_dens : float
        Number density of background stars in the direction of the object of 
        interest, in arcsec^-2.
    sep : float or numpy 1d array
        Separation of the point sources with respect to central star, in arcsec.
    n_bkg : int, opt
        Number of point sources in the field, and for which the separation is
        provided.
    verbose: bool, opt
        Whether to print the probabilities for 0 to n_bkg point sources.
    full_output: bool, opt
        Whether to also return probabilities of 0 to n_bkg-1 point sources

    Returns
    -------
    proba : float
        Probability between 0% and 100%.
    [probas : np 1d array] if full_output is True
        Probabilities of getting 0 to n_bkg-1 point sources

    """
    
    if n_bkg < 1 or not isinstance(n_bkg,int):
        raise TypeError("n_bkg should be a strictly positive integer.")
    
    if not isinstance(sep, float):
        if isinstance(sep, np.ndarray):
            if sep.ndim!=1 or sep.shape[0]!=n_bkg:
                raise TypeError("if sep is a np array, its len should be n_bkg")
            else:
                sep = np.amax(sep)
        else:
            raise TypeError("sep can only be a float or a np 1d array")

    B = np.pi*sep**2 
    probas = np.zeros(n_bkg)
    for i in range(n_bkg):
        probas[i] = np.exp(-n_dens*B)*(n_dens*B)**i/float(factorial(i))
        if verbose:
            msg = "Proba of having {:.0f} bkg star in a disk of "
            msg += "{:.1}'' radius: {:.1f}%"
            print(msg.format(i,sep,probas[i]))

    proba = 1-np.sum(probas)
    if verbose:
        msg = "Proba of having {:.0f} bkg star or more in a disk of "
        msg += "{:.1}'' radius: {:.1f}%"
        print(msg.format(n_bkg,sep,proba))

    if full_output:
        return proba, probas
    else:
        return proba


#! /usr/bin/env python

"""
Distance and correlation between images.
"""

__author__ = 'Carlos Alberto Gomez Gonzalez; V. Christiaens'
__all__ = ['cube_distance',
           'spectral_correlation']

from astropy.stats import gaussian_fwhm_to_sigma
import numpy as np
import scipy.stats
from scipy.optimize import curve_fit
from matplotlib import pyplot as plt
try:
    # for skimage version >= '0.16' use skimage.metrics.structural_similarity
    from skimage.metrics import structural_similarity as ssim
except:
    # before skimage version '0.16' use skimage.measure.compare_ssim
    from skimage.measure import compare_ssim as ssim
from ..var import get_annulus_segments, get_circle
from ..conf import vip_figsize


def cube_distance(array, frame, mode='full', dist='sad', inradius=None,
                  width=None, mask=None, plot=True):
    """ Computes the distance (or similarity) between frames in a cube, using
    one as the reference (it can be either a frame from the same cube or a
    separate 2d array). Depending on the mode, the whole image can be used,
    or just the pixels in a given annulus. The criteria used are:
    - the Manhattan distance (SAD or sum of absolute differences),
    - the Euclidean distance (square root of the sum of the squared differences),
    - the Mean Squared Error,
    - the Spearman correlation coefficient,
    - the Pearson correlation coefficient,
    - the Structural Similarity Index (SSIM).

    The SAD, MSE and Ecuclidean criteria are dissimilarity criteria, which
    means that 0 is perfect similarity.
    The Spearman and Pearson correlation coefficients, vary between -1 and +1
    with 0 implying no correlation. Correlations of -1 or +1 imply an exact
    linear relationship.
    The Structural Similarity Index was proposed by Wang et al. 2004.
    (http://www.cns.nyu.edu/pub/eero/wang03-reprint.pdf)
    SSIM varies between -1 and 1, where 1 means perfect similarity. SSIM
    attempts to model the perceived change in the structural information of the
    image. The mean SSIM is reported.

    Parameters
    ----------
    array : numpy ndarray
        Input cube or 3d array.
    frame : int, 2d array or None
        Reference frame in the cube or 2d array. If None, will take the median
        frame of the 3d array.
    mode : {'full','annulus', 'mask'}, string optional
        Whether to use the full frames, a centered annulus or a provided mask.
    dist : {'sad','euclidean','mse','pearson','spearman', 'ssim'}, str optional
        Which criterion to use.
    inradius : None or int, optional
        The inner radius when mode is 'annulus'.
    width : None or int, optional
        The width when mode is 'annulus'.
    mask: 2d array, optional
        If mode is 'mask', this is the mask within which the metrics is
        calculated in the images.
    plot : bool, optional
        Whether to plot the distances or not.

    Returns
    -------
    lista : numpy ndarray
        1d array of distances for each frame wrt the reference one.

    """
    if array.ndim != 3:
        raise TypeError('The input array is not a cube or 3d array')
    lista = []
    n = array.shape[0]
    if isinstance(frame, int):
        frame_ref = array[frame]
    elif isinstance(frame, np.ndarray):
        frame_ref = frame
    elif frame is None:
        frame_ref = np.median(array, axis=0)
    else:
        raise TypeError('Input ref frame format not recognized')
    if mode == 'full':
        pass
    elif mode == 'annulus':
        if inradius is None:
            raise ValueError('`Inradius` has not been set')
        if width is None:
            raise ValueError('`Width` has not been set')
        frame_ref = get_annulus_segments(frame_ref, inradius, width,
                                         mode="val")[0]
    elif mode == 'mask':
        if mask is None:
            raise ValueError('mask has not been set')
        frame_ref = frame_ref[np.where(mask)]
    else:
        raise TypeError('Mode not recognized or missing parameters')

    for i in range(n):
        if mode == 'full':
            framei = array[i]
        elif mode == 'annulus':
            framei = get_annulus_segments(array[i], inradius, width,
                                          mode="val")[0]
        elif mode == 'mask':
            framei = array[i][np.where(mask)]
        if dist == 'sad':
            lista.append(np.sum(abs(frame_ref - framei)))
        elif dist == 'euclidean':
            lista.append(np.sqrt(np.sum((frame_ref - framei)**2)))
        elif dist == 'mse':
            lista.append((np.sum((frame_ref - framei)**2))/len(frame_ref))
        elif dist == 'pearson':
            pears, _ = scipy.stats.pearsonr(frame_ref.ravel(), framei.ravel())
            lista.append(pears)
        elif dist == 'spearman':
            spear, _ = scipy.stats.spearmanr(frame_ref.ravel(), framei.ravel())
            lista.append(spear)
        elif dist == 'ssim':
            mean_ssim = ssim(frame_ref, framei, win_size=7,
                             data_range=frame_ref.max() - frame_ref.min(),
                             gaussian_weights=True, sigma=1.5,
                             use_sample_covariance=True)
            lista.append(mean_ssim)
        else:
            raise ValueError('Distance not recognized')
    lista = np.array(lista)

    median_cor = np.median(lista)
    mean_cor = np.mean(lista)
    if plot:
        _, ax = plt.subplots(figsize=vip_figsize)

        if isinstance(frame, int):
            ax.vlines(frame, ymin=np.nanmin(lista), ymax=np.nanmax(lista),
                      colors='green', linestyles='dashed', lw=2, alpha=0.8,
                      label='Frame '+str(frame))
        ax.hlines(median_cor, xmin=-1, xmax=n+1, colors='purple', alpha=0.3,
                  linestyles='dashed', label='Median value : '+str(median_cor))
        ax.hlines(mean_cor, xmin=-1, xmax=n+1, colors='green', alpha=0.3,
                  linestyles='dashed', label='Mean value : '+str(mean_cor))

        x = range(len(lista))
        ax.plot(x, lista, '-', alpha=0.6, color='#1f77b4')
        ax.plot(x, lista, 'o', alpha=0.4, color='#1f77b4')
        plt.xlabel('Frame number')
        if dist == 'sad':
            plt.ylabel('SAD - Manhattan distance')
        elif dist == 'euclidean':
            plt.ylabel('Euclidean distance')
        elif dist == 'pearson':
            plt.ylabel('Pearson correlation coefficient')
        elif dist == 'spearman':
            plt.ylabel('Spearman correlation coefficient')
        elif dist == 'mse':
            plt.ylabel('Mean squared error')
        elif dist == 'ssim':
            plt.ylabel('Structural Similarity Index')

        plt.xlim(xmin=-1, xmax=n+1)
        plt.minorticks_on()
        plt.legend(fancybox=True, framealpha=0.5, fontsize=12, loc='best')
        plt.grid(which='major', alpha=0.2)

    return lista



def spectral_correlation(array, ann_width=2, r_in=1, r_out=None, pl_xy=None,
                         mask_r=4, fwhm=4, sp_fwhm_guess=3,full_output=False):
    """ Computes the spectral correlation between (post-processed) IFS frames, 
    as a function of radius, implemented as Eq. 7 of Greco & Brandt 2017. This 
    is a crucial step for an unbias fit of a measured IFS spectrum to either 
    synthetic or template spectra.
    
    Parameters
    ----------
    array : numpy ndarray
        Input cube or 3d array, of dimensions n_ch x n_y x n_x; where n_y and 
        n_x should be odd values (star should be centered on central pixel).
    ann_width : int, optional
        Width in pixels of the concentric annuli used to compute the spectral 
        correlation as a function of radial separation. Greco & Brandt 2017 
        noted no significant differences for annuli between 1 and 3 pixels 
        width on GPI data.
    r_in: int, optional
        Innermost radius where the spectral correlation starts to be computed.
    r_out: int, optional
        Outermost radius where the spectral correlation is computed.If left as 
        None, it will automatically be computed up to the edge of the frame. 
    pl_xy: tuple of tuples of 2 floats, optional
        x,y coordiantes of all companions present in the images.
        If provided, a circle centered on the location of each
        companion will be masked out for the spectral correlation computation.
    mask_r: float, optional
        if pl_xy is provided, this should also be provided. Size of the 
        aperture around each companion (in terms of fwhm) that is discarded to 
        not bias the spectral correlation computation.
    fwhm: float, optional
        if pl_xy is provided, this should also be provided. By default we  
        consider a 2FWHM aperture mask around each companion to not bias the 
        spectral correlation computation.
    sp_fwhm_guess: float, optional
        Initial guess on the spectral FWHM of all channels.
    full_output: bool, opt
        Whether to also output the fitted spectral FWHM for each channel.
    Note: radii that are skipped will be filled with zeros in the output cube.

    Returns
    -------
    sp_corr : numpy ndarray
        3d array of spectral correlation, as a function of radius with 
        dimensions: n_r x n_ch x n_ch, where n_r = min((n_y-1)/2,(n_x-1)/2)
        Starts at r = 1 (not r=0) px.
    sp_fwhm: numpy ndarray
        (if full_output is True) 2d array containing the spectral fwhm at each 
        radius, for each spectral channel. Dims: n_r x n_ch
        
    """

    if not isinstance(ann_width,int) or not isinstance(r_in,int):
        raise TypeError("Inputs should be integers")

    if array.ndim != 3:
        raise TypeError("Input array should be 3D.")
        
    n_ch, n_y, n_x = array.shape
    n_r = min((n_y-1)/2.,(n_x-1)/2.)
    if n_r%1:
        raise TypeError("Input array y and x dimensions should be odd")
    
    if r_out is None:
        r_out = n_r

    test_rads = np.arange(r_in-1,r_out-1)
    n_rad = int(np.floor(test_rads.shape[0]/ann_width))
    
    #n_rad = int(np.ceil(n_r/ann_width)) # effective number of annuli probed
    
    sp_corr = np.zeros([int(n_r),n_ch,n_ch])
    if full_output:
        sp_fwhm = np.zeros([int(n_r),n_ch])
        def gauss_1fp(x, *p):
            sigma = p[0]*gaussian_fwhm_to_sigma
            return np.exp(-x**2/(2.*sigma**2))
    mask_final = np.zeros_like(array[0])

    if pl_xy is not None:
        mask = np.ones_like(array[0])
        for i in range(len(pl_xy)):
            if not isinstance(pl_xy[i], tuple):
                raise TypeError("Format of companions coordinates incorrect")
            mask_i = get_circle(mask, radius=mask_r*fwhm, cy=pl_xy[i][1], cx=pl_xy[i][0], mode="mask")
            mask_final[np.where(mask_i)] = 1

    for ann in range(n_rad):
        inner_radius = r_in+ (ann * ann_width)
        indices = get_annulus_segments(array[0], inner_radius, ann_width)
        yy = indices[0][0]
        xx = indices[0][1]
        yy_final = [yy[i] for i in range(len(indices[0][0])) if not mask_final[yy[i],xx[i]]]
        xx_final = [xx[i] for i in range(len(indices[0][0])) if not mask_final[yy[i],xx[i]]]
        matrix = array[:, yy_final, xx_final]  # shape (z, npx_annsegm)
        for zi in range(n_ch):
            for zj in range(n_ch):
                num = np.nanmean(matrix[zi]*matrix[zj])
                denom = np.sqrt(np.nanmean(matrix[zi]*matrix[zi])* \
                                np.nanmean(matrix[zj]*matrix[zj]))
                sp_corr[r_in+ann*ann_width:r_in+(ann+1)*ann_width,zi,zj] = num/denom
            if full_output:
                p0 = (sp_fwhm_guess,)
                x = np.arange(n_ch)-zi
                y = sp_corr[r_in+ann*ann_width,zi]# norm y
                y = y-np.amin(y)
                y = y/np.amax(y)
                coeff, var_matrix = curve_fit(gauss_1fp, x, y, p0=p0)
                sp_fwhm[r_in+ann*ann_width:r_in+(ann+1)*ann_width,zi] = coeff[0]

                
    if full_output:
        return sp_corr, sp_fwhm
    else:
        return sp_corr
    


#! /usr/bin/env python

"""
Functions useful for spectral fitting of companions, and model interpolation.
"""

__author__ = 'Valentin Christiaens'
__all__ = ['make_model_from_params',
           'make_resampled_models',
           'resample_model',
           'interpolate_model']

import numpy as np
import astropy.constants as con
from astropy.convolution import Gaussian1DKernel, convolve_fft
from astropy.stats import gaussian_fwhm_to_sigma
import itertools
from scipy.interpolate import InterpolatedUnivariateSpline
import pandas as pd
import pdb
from scipy.ndimage import map_coordinates
from ..fits import open_fits
from .utils_spec import (convert_F_units, blackbody, find_nearest, extinction,
                         inject_em_line)


def make_model_from_params(params, labels, grid_param_list, dist, lbda_obs=None, 
                           model_grid=None, model_reader=None, em_lines={}, 
                           em_grid={}, dlbda_obs=None, instru_fwhm=None, 
                           instru_idx=None, filter_reader=None, AV_bef_bb=False,
                           units_obs='si', units_mod='si', interp_order=1):
    """
    Routine to make the model from input parameters.
    
        Parameters
    ----------
    params : tuple
        Set of models parameters for which the model grid has to be 
        interpolated.
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state, that is:
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - then the planet photometric radius 'R', in Jupiter radius
        - (optionally) the flux of emission lines (labels should match those in
        the em_lines dictionary), in units of the model spectrum (times mu)
        - (optionally) the optical extinction 'Av', in mag
        - (optionally) the ratio of total to selective optical extinction 'Rv'
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb
        contribution.
    grid_param_list : list of 1d numpy arrays/lists OR None
        - If list, should contain list/numpy 1d arrays with available grid of 
        model parameters. 
        - Set to None for a pure n-blackbody fit, n=1,2,...
        - Note1: model grids should not contain grids on radius and Av, but 
        these should still be passed in initial_state (Av optional).
        - Note2: for a combined grid model + black body, just provide
        the grid parameter list here, and provide values for 'Tbbn' and 'Rbbn'
        in initial_state, labels and bounds.
    dist :  float
        Distance in parsec, used for flux scaling of the models.
    lbda_obs : numpy 1d ndarray or list, opt
        Wavelength of observed spectrum. If provided, the model spectrum will 
        be resampled to match lbda_obs. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs). 
    model_grid : numpy N-d array, optional
        If provided, should contain the grid of model spectra for each
        free parameter of the given grid. I.e. for a grid of n_T values of Teff 
        and n_g values of Logg, the numpy array should be n_T x n_g x n_ch x 2, 
        where n_ch is the number of wavelengths for the observed spectrum,
        and the last 2 dims are for wavelength and fluxes respectively.
        If provided, takes precedence over model_name/model_reader.
    model_reader : python routine, opt
        External routine that reads a model file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains model values. See example routine in interpolate_model() 
        description.
    em_lines: dictionary, opt
        Dictionary of emission lines to be added on top of the model spectrum.
        Each dict entry should be the name of the line, assigned to a tuple of
        3 values: 1) the wavelength (in mu), 2) a string indicating whether line 
        intensity is expressed in flux ('F'), luminosity ('L') or log(L/LSun) 
        ("LogL"), and 3) the latter quantity. The intensity of the emission 
        lines can be sampled by MCMC, in that case the last element of the 
        tuple can be set to None. If not to be sampled, a value for the 
        intensity should be provided (in the same system of units as the model 
        spectra, multiplied by mu). Examples:
        em_lines = {'BrG':(2.1667,'F',263)};
        em_lines = {'BrG':(2.1667,'LogL',-5.1)}
    em_grid: dictionary pointing to lists, opt
        Dictionary where each entry corresponds to an emission line and points
        to a list of values to inject for emission line fluxes. For computation 
        efficiency, interpolation will be performed between the points of this 
        grid during the MCMC sampling. Dict entries should match labels and 
        em_lines.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_fwhm : float or list, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band
        filter FWHM if relevant.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default format assumed for the files:
        - first row containing header
        - starting from 2nd row: 1st column: WL in mu, 2nd column: transmission
        Note: files should all have the same format and wavelength units.
    AV_bef_bb: bool, optional
        If both extinction and an extra bb component are free parameters, 
        whether to apply extinction before adding the BB component (e.g. 
        extinction mostly from circumplanetary dust) or after the BB component
        (e.g. mostly insterstellar extinction).
    units_obs : str, opt {'si','cgs','jy'}
        Units of observed spectrum. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu 
        or 'jy' for janskys.
    units_mod: str, opt {'si','cgs','jy'}
        Units of the model. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu or 'jy'
        for janskys. If different to units_obs, the spectrum units will be 
        converted.
    interp_order: int, opt, {-1,0,1} 
        Interpolation mode for model interpolation.
        -1: log interpolation (i.e. linear interpolatlion on log(Flux))
        0: nearest neighbour model.
        1: Order 1 spline interpolation.
        
    Returns
    -------
    out: numpy array
        The model wavelength and spectrum
    """
    
    
    if 'Tbb1' in labels and grid_param_list is None and lbda_obs is None:
        raise ValueError("lbda_obs should be provided because there is no grid")
      
    if grid_param_list is None:
        lbda_mod = lbda_obs
        spec_mod = np.zeros_like(lbda_obs)
    else:
        npar_grid = len(grid_param_list)
        params_grid = [params[i] for i in range(npar_grid)]
        params_grid = tuple(params_grid)        
        if len(em_grid) == 0:
            p_em_grid = None
        else:
            p_em_grid = {}
            for key, _ in em_grid.items():
                j = labels.index(key)
                p_em_grid[key] = params[j]
        # interpolate model to requested parameters
        lbda_mod, spec_mod = interpolate_model(params_grid, grid_param_list, 
                                               p_em_grid, em_grid, em_lines,
                                               labels, model_grid, 
                                               model_reader, interp_order)
        
        # resample to lbda_obs if needed
        if lbda_obs is not None:
            cond = False
            if len(lbda_obs) != len(lbda_mod):
                cond = True
            elif not np.allclose(lbda_obs, lbda_mod):
                cond = True
            if cond:
                lbda_mod, spec_mod = resample_model(lbda_obs, lbda_mod, spec_mod, 
                                                    dlbda_obs, instru_fwhm, 
                                                    instru_idx, filter_reader)
            
        # convert model to same units as observed spectrum if necessary
        if units_mod != units_obs:
            spec_mod = convert_F_units(spec_mod, lbda_mod, in_unit=units_mod, 
                                       out_unit=units_obs)

        # scale by (R/dist)**2
        idx_R = labels.index("R")
        dilut_fac = ((params[idx_R]*con.R_jup.value)/(dist*con.pc.value))**2
        spec_mod *= dilut_fac

        
    # apply extinction if requested
    if 'Av' in labels and AV_bef_bb:
        ## so far only assume Cardelli extinction law
        idx_AV = labels.index("Av")
        if 'Rv' in labels:
            idx_RV = labels.index("Rv")
            RV = params[idx_RV]
        else:
            RV=3.1
        extinc_curve = extinction(lbda_mod, params[idx_AV], RV)
        flux_ratio_ext = np.power(10.,-extinc_curve/2.5)
        spec_mod *= flux_ratio_ext
        ## TBD: add more options

    # add n blackbody component(s) if requested 
    if 'Tbb1' in labels:
         n_bb = 0
         for label in labels:
             if 'Tbb' in label:
                 n_bb+=1
         idx_Tbb1 = labels.index("Tbb1")
         Rj = con.R_jup.value
         pc = con.pc.value
         for ii in range(n_bb):
             idx = ii*2
             Omega = np.pi*((params[idx_Tbb1+idx+1]*Rj)/(dist*pc))**2
             bb = Omega*blackbody(lbda_mod, params[idx_Tbb1+idx])
             spec_mod += bb

    # apply extinction if requested
    if 'Av' in labels and not AV_bef_bb:
        ## so far only assume Cardelli extinction law
        idx_AV = labels.index("Av")
        if 'Rv' in labels:
            idx_RV = labels.index("Rv")
            RV = params[idx_RV]
        else:
            RV=3.1
        extinc_curve = extinction(lbda_mod, params[idx_AV], RV)
        flux_ratio_ext = np.power(10.,-extinc_curve/2.5)
        spec_mod *= flux_ratio_ext
        ## TBD: add more options                
        
    return lbda_mod, spec_mod


def make_resampled_models(lbda_obs, grid_param_list, model_grid=None,
                          model_reader=None, em_lines={}, em_grid=None, 
                          dlbda_obs=None, instru_fwhm=None, instru_idx=None, 
                          filter_reader=None, interp_nonexist=True):
    """
    Returns a cube of models after convolution and resampling as in the 
    observations.
    
    Parameters:
    -----------   
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).
    grid_param_list : list of 1d numpy arrays/lists
        Should contain list/numpy 1d arrays with available grid of model 
        parameters. Note: model grids shouldn't contain grids on radius and Av.
    model_grid: list of 1d numpy arrays, or list of lists.
        Available grid of model parameters (should only contain the parameter
        values, not the models themselves). The latter will be loaded. 
        Important: 1) Make sure the bounds are within the model grid to avoid 
        extrapolation. 2) All keywords that are neither 'R', 'Av' nor 'M' will 
        be considered model grid parameters.
        length of params, with the same order. OR '1bb' or '2bb' for black-body
        models. In that case the model will be created on the fly at each 
        iteration using 1 or 2 Planck functions respectively. There are 2 params 
        for each Planck function: Teff and radius.
    model_reader : python routine
        External routine that reads a model file, converts it to required 
        units and returns a 2D numpy array, where the first column corresponds
        to wavelengths, and the second contains model values. Example below.
    em_lines: dictionary, opt
        Dictionary of emission lines to be added on top of the model spectrum.
        Each dict entry should be the name of the line, assigned to a tuple of
        3 values: 1) the wavelength (in mu), 2) a string indicating whether line 
        intensity is expressed in flux ('F'), luminosity ('L') or log(L/LSun) 
        ("LogL"), and 3) the latter quantity. The intensity of the emission 
        lines can be sampled by MCMC, in that case the last element of the 
        tuple can be set to None. If not to be sampled, a value for the 
        intensity should be provided (in the same system of units as the model 
        spectra, multiplied by mu). Example:
        em_lines = {'BrG':(2.1667,'F',263)}
    em_grid: dictionary pointing to lists, opt
        Dictionary where each entry corresponds to an emission line and points
        to a list of values to inject for emission line fluxes. For computation 
        efficiency, interpolation will be performed between the points of this 
        grid during the MCMC sampling. Dict entries should match labels and 
        em_lines. Note: length of this dictionary can be different of em_lines;
        i.e. if a line is in em_lines but not in em_grid, it will not be 
        considered an MCMC parameter.
    lbda_mod : numpy 1d ndarray or list
        Wavelength of tested model. Should have a wider wavelength extent than 
        the observed spectrum.
    spec_mod : numpy 1d ndarray
        Model spectrum. It does not require the same wavelength sampling as the
        observed spectrum. If higher spectral resolution, it will be convolved
        with the instrumental spectral psf (if instru_fwhm is provided) and 
        then binned to the same sampling. If lower spectral resolution, a 
        linear interpolation is performed to infer the value at the observed 
        spectrum wavelength sampling.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_fwhm : float or list, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band
        filter FWHM if relevant.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default format assumed for the files:
        - first row containing header
        - starting from 2nd row: 1st column: WL in mu, 2nd column: transmission
        Note: files should all have the same format and wavelength units.
    interp_nonexist: bool, opt
        Whether to interpolate if models do not exist, based on closest model(s)
        
    Returns
    -------
    resamp_mod: 1d numpy array
        Grid of model spectra resampled at wavelengths matching the observed 
        spectrum.
    """
    n_params = len(grid_param_list)
    n_mods = len(grid_param_list[0])
    dims = [len(grid_param_list[0])]
    
    if n_params>1:
        for pp in range(1,n_params):
            n_mods *= len(grid_param_list[pp])
            dims.append(len(grid_param_list[pp]))

    if em_grid is None:
        n_em = 0
        final_dims = dims+[len(lbda_obs)]+[2]
    else:
        n_em = len(em_grid)
        n_em_mods = 1
        dims_em = []
        for key, _ in em_grid.items():
            n_em_mods *= len(em_grid[key])
            dims_em.append(len(em_grid[key]))
        final_dims = dims+dims_em+[len(lbda_obs)]+[2]
        dims_em = tuple(dims_em)
     
    final_dims = tuple(final_dims)
    dims = tuple(dims)
    resamp_mod = []
    
    # Loop on all models whose parameters are provided in model grid
    for nn in range(n_mods):            
        if model_grid is not None:
            indices = []
            idx = np.unravel_index(nn,dims)
            for pp in range(n_params):
                indices.append(idx[pp])
            indices = tuple(indices)
            tmp = model_grid[indices]
            lbda_mod = tmp[:,0]
            spec_mod = tmp[:,1]             
        else:
            params_tmp = []
            idx = np.unravel_index(nn,dims)
            for pp in range(n_params):
                params_tmp.append(grid_param_list[pp][idx[pp]])  
            try:
                lbda_mod, spec_mod = model_reader(params_tmp)
                if np.sum(np.isnan(spec_mod))>0:
                    print("There are nan values in spec for params: ")
                    pdb.set_trace()
            except:
                msg= "Model does not exist for param combination ({})"
                print(msg.format(params_tmp))
                print("Press c if you wish to interpolate that model from neighbours")
                pdb.set_trace()
                if interp_nonexist:
                    # find for which dimension the model doesn't exist;
                    for qq in range(n_params):
                        interp_params1=[]
                        interp_params2=[]
                        for pp in range(n_params):
                            if pp == qq:
                                try:
                                    interp_params1.append(grid_param_list[pp][idx[pp]-1])
                                    interp_params2.append(grid_param_list[pp][idx[pp]+1])
                                except:
                                    continue
                            else:
                                interp_params1.append(grid_param_list[pp][idx[pp]])
                                interp_params2.append(grid_param_list[pp][idx[pp]])
                        try:
                            lbda_mod1, spec_mod1 = model_reader(interp_params1)
                            lbda_mod2, spec_mod2 = model_reader(interp_params2)
                            lbda_mod = np.mean([lbda_mod1,lbda_mod2],axis=0)
                            spec_mod = np.mean([spec_mod1,spec_mod2],axis=0)
                            msg= "Model was interpolated based on models: {} "
                            msg+="and {}"
                            print(msg.format(interp_params1,interp_params2))
                            break
                        except:
                            pass
                        if qq == n_params-1:
                            msg = "Impossible to interpolate model!" 
                            msg += "Consider reducing bounds."
                            raise ValueError(msg)
                else:
                    msg = "Model interpolation not allowed for non existing " 
                    msg += "models in the grid."
                    raise ValueError(msg)                    
                    
        # inject emission lines if any
        if n_em > 0:
            flux_grids = []
            wls = []
            widths = []
            for key, flux_grid in em_grid.items():
                flux_grids.append(flux_grid)
                wls.append(em_lines[key][0])
                widths.append(em_lines[key][2])
            # recursively inject em lines
            for fluxes in itertools.product(*flux_grids):
                for ff, flux in enumerate(fluxes):
                    spec_mod = inject_em_line(wls[ff], flux, lbda_mod, spec_mod,
                                              widths[ff])                
                # interpolate OR convolve+bin model spectrum if required
                if len(lbda_obs) != len(lbda_mod):
                    res = resample_model(lbda_obs, lbda_mod, spec_mod, 
                                         dlbda_obs, instru_fwhm, instru_idx, 
                                         filter_reader)
                elif not np.allclose(lbda_obs, lbda_mod):
                    res = resample_model(lbda_obs, lbda_mod, spec_mod, 
                                         dlbda_obs, instru_fwhm, instru_idx, 
                                         filter_reader)
                else:
                    res = np.array([lbda_obs, spec_mod])
                    
                resamp_mod.append(res)                
        else:
                    
            # interpolate OR convolve+bin model spectrum if not same sampling
            if len(lbda_obs) != len(lbda_mod):
                res = resample_model(lbda_obs, lbda_mod, spec_mod, dlbda_obs, 
                                     instru_fwhm, instru_idx, filter_reader)
            elif not np.allclose(lbda_obs, lbda_mod):
                res = resample_model(lbda_obs, lbda_mod, spec_mod, dlbda_obs, 
                                     instru_fwhm, instru_idx, filter_reader)
            else:
                res = np.array([lbda_obs, spec_mod])
                
            resamp_mod.append(res)

    resamp_mod = np.array(resamp_mod)
    resamp_mod = np.swapaxes(resamp_mod,-1,-2)
       
    return resamp_mod.reshape(final_dims)

 
def resample_model(lbda_obs, lbda_mod, spec_mod, dlbda_obs=None, 
                   instru_fwhm=None, instru_idx=None, filter_reader=None,
                   no_constraint=False, verbose=False):
    """
    Convolve, interpolate and resample a model spectrum to match observed 
    spectrum.

    Parameters:
    -----------
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).
    lbda_mod : numpy 1d ndarray or list
        Wavelength of tested model. Should have a wider wavelength extent than 
        the observed spectrum.
    spec_mod : numpy 1d ndarray
        Model spectrum. It does not require the same wavelength sampling as the
        observed spectrum. If higher spectral resolution, it will be convolved
        with the instrumental spectral psf (if instru_fwhm is provided) and 
        then binned to the same sampling. If lower spectral resolution, a 
        linear interpolation is performed to infer the value at the observed 
        spectrum wavelength sampling.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_fwhm : float or list, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band
        filter FWHM if relevant.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default format assumed for the files:
        - first row containing header
        - starting from 2nd row: 1st column: WL in mu, 2nd column: transmission
        Note: files should all have the same format and wavelength units.
    no_constraint: bool, optional
        If set to True, will not use 'floor' and 'ceil' constraints when
        cropping the model wavelength ranges, i.e. faces the risk of 
        extrapolation. May be useful, if the bounds of the wavelength ranges 
        are known to match exactly. 
    verbose: bool, optional
        Whether to print more information during resampling.
        
    Returns
    -------
    lbda_obs, spec_mod_res: 2x 1d numpy array
        Observed lambdas, and resampled model spectrum (at those lambdas)
    
    """
    
    def _default_file_reader(filter_name):
        """
        Default file reader if no filter file reader is provided.
        """
        filt_table = pd.read_csv(filter_name, sep=' ', header=0, 
                                 skipinitialspace=True)
        keys = filt_table.keys()
        lbda_filt = np.array(filt_table[keys[0]])
        if '(AA)' in keys[0]:
            lbda_filt /=10000
        elif '(mu)' in keys[0]:
            pass
        elif '(nm)' in keys[0]:
            lbda_filt /=10000
        else:
            raise ValueError('Wavelength unit not recognised in filter file')
        trans = np.array(filt_table[keys[1]])
        return lbda_filt, trans      
                    
    n_ch = len(lbda_obs)
    spec_mod_res = np.zeros_like(lbda_obs)
    
    if dlbda_obs is None:
        # this is only to trim out useless WL ranges, hence significantly 
        # improving speed for large (>1M pts) models (e.g. BT-SETTL).
        # 0.3 factor to consider possible broad-band filters.
        dlbda_obs1 = [min(0.3*lbda_obs[0],lbda_obs[1]-lbda_obs[0])]
        dlbda_obs2 = [(lbda_obs[i+2]-lbda_obs[i])/2 for i in range(n_ch-2)]
        dlbda_obs3 = [min(0.3*lbda_obs[-1],lbda_obs[-1]-lbda_obs[-2])]
        dlbda_obs = np.array(dlbda_obs1+dlbda_obs2+dlbda_obs3)

    if verbose:
        print("checking whether WL samplings are the same for obs and model")        
        
    if isinstance(instru_fwhm, float) or isinstance(instru_fwhm, int):
        instru_fwhm = [instru_fwhm]
    cond = False
    if len(lbda_obs) != len(lbda_mod):
        cond = True
    elif not np.allclose(lbda_obs, lbda_mod):
        cond = True
    if cond:
        lbda_min = lbda_obs[0]-dlbda_obs[0]
        lbda_max = lbda_obs[-1]+dlbda_obs[-1]
        try:
            lbda_min_tmp = min(lbda_min,lbda_obs[0]-3*np.amax(instru_fwhm)/1000)
            lbda_max_tmp = max(lbda_max,lbda_obs[-1]+3*np.amax(instru_fwhm)/1000)
            if lbda_min_tmp > 0:
                lbda_min = lbda_min_tmp
                lbda_max = lbda_max_tmp
        except:
            pass
            
        if no_constraint:
            idx_ini = find_nearest(lbda_mod, lbda_min)
            idx_fin = find_nearest(lbda_mod, lbda_max)
        else:
            idx_ini = find_nearest(lbda_mod, lbda_min, 
                                   constraint='floor')
            idx_fin = find_nearest(lbda_mod, lbda_max, 
                                   constraint='ceil')

        lbda_mod = lbda_mod[idx_ini:idx_fin+1]
        spec_mod = spec_mod[idx_ini:idx_fin+1]
        
        
    nmod = lbda_mod.shape[0]
    ## compute the wavelength sampling of the model
    dlbda_mod1 = [lbda_mod[1]-lbda_mod[0]]
    dlbda_mod2 = [(lbda_mod[i+1]-lbda_mod[i-1])/2 for i in range(1,nmod-1)]
    dlbda_mod3 = [lbda_mod[-1]-lbda_mod[-2]]
    dlbda_mod = np.array(dlbda_mod1+dlbda_mod2+dlbda_mod3)


    if verbose:
        print("testing whether observed spectral res could be > than model's ")
        print("(in at least parts of the spectrum)")
    dlbda_obs_min = np.amin(dlbda_obs)
    idx_obs_min = np.argmin(dlbda_obs)
    idx_near = find_nearest(lbda_mod, lbda_obs[idx_obs_min])
    dlbda_mod_tmp = (lbda_mod[idx_near+1]-lbda_mod[idx_near-1])/2
    do_interp = np.zeros(n_ch, dtype='int32')
    
    if dlbda_mod_tmp>dlbda_obs_min and dlbda_obs_min > 0:
        if verbose:
            print("checking where obs spec res is < or > than model's")
        ## check where obs spec res is < or > than model's
        nchunks_i = 0
        for ll in range(n_ch):
            idx_near = find_nearest(lbda_mod, lbda_obs[ll])
            do_interp[ll] = (dlbda_obs[ll] < dlbda_mod[idx_near])
            if ll > 0:
                if do_interp[ll] and not do_interp[ll-1]:
                    nchunks_i+=1
            elif do_interp[ll]:
                nchunks_i=1
            
    ## interpolate model if the observed spectrum has higher resolution 
    ## and is monotonically increasing
    if np.sum(do_interp) and dlbda_obs_min > 0:
        if verbose:
            print("interpolating model where obs spectrum has higher res")    
        idx_0=0
        for nc in range(nchunks_i):
            idx_1 = np.argmax(do_interp[idx_0:])+idx_0
            idx_0 = np.argmin(do_interp[idx_1:])+idx_1
            if idx_0==idx_1:
                idx_0=-1
                if nc != nchunks_i-1:
                    pdb.set_trace() # should not happen
            idx_ini = find_nearest(lbda_mod,lbda_obs[idx_1],
                                   constraint='floor')
            idx_fin = find_nearest(lbda_mod,lbda_obs[idx_0],
                                   constraint='ceil')


            spl = InterpolatedUnivariateSpline(lbda_mod[idx_ini:idx_fin], 
                                               spec_mod[idx_ini:idx_fin],
                                                   k=min(3,idx_fin-idx_ini-1))
            spec_mod_res[idx_1:idx_0] = spl(lbda_obs[idx_1:idx_0])

            
    ## convolve+bin where the model spectrum has higher resolution (most likely)
    if np.sum(do_interp) < n_ch or dlbda_obs_min > 0:
    # Note: if dlbda_obs_min < 0, it means several instruments are used with 
    # overlapping WL ranges. instru_fwhm should be provided!
        if instru_fwhm is None:
            msg = "Warning! No spectral FWHM nor filter file provided"
            msg+= " => binning without convolution"
            print(msg)
            for ll, lbda in enumerate(lbda_obs):
                mid_lbda_f = lbda_obs-dlbda_obs/2.
                mid_lbda_l = lbda_obs+dlbda_obs/2.
                i_f = find_nearest(lbda_mod,
                                   mid_lbda_f[ll])
                i_l = find_nearest(lbda_mod,
                                   mid_lbda_l[ll])
                spec_mod_res[ll] = np.mean(spec_mod[i_f:i_l+1])
        else:
            if verbose:
                print("convolving+binning where model spectrum has higher res")   
            if isinstance(instru_idx, list):
                instru_idx = np.array(instru_idx)
            elif not isinstance(instru_idx, np.ndarray):
                instru_idx = np.array([1]*n_ch)
    
            for i in range(1,len(instru_fwhm)+1):
                if isinstance(instru_fwhm[i-1], (float,int)):
                    ifwhm = instru_fwhm[i-1]/(1000*np.mean(dlbda_mod))
                    gau_ker = Gaussian1DKernel(stddev=ifwhm*gaussian_fwhm_to_sigma)
                    spec_mod_conv = convolve_fft(spec_mod, gau_ker)
                    tmp = np.zeros_like(lbda_obs[np.where(instru_idx==i)])
                    for ll, lbda in enumerate(lbda_obs[np.where(instru_idx==i)]):
                        mid_lbda_f = lbda_obs-dlbda_obs/2.
                        mid_lbda_l = lbda_obs+dlbda_obs/2.
                        i_f = find_nearest(lbda_mod,
                                           mid_lbda_f[np.where(instru_idx==i)][ll])
                        i_l = find_nearest(lbda_mod,
                                           mid_lbda_l[np.where(instru_idx==i)][ll])
                        tmp[ll] = np.mean(spec_mod_conv[i_f:i_l+1])
                    spec_mod_res[np.where(instru_idx==i)] = tmp  
                elif isinstance(instru_fwhm[i-1], str):
                    if filter_reader is not None:
                        lbda_filt, trans = filter_reader(instru_fwhm[i-1])
                    else:
                        lbda_filt, trans = _default_file_reader(instru_fwhm[i-1])
                    idx_ini = find_nearest(lbda_mod, lbda_filt[0], 
                                           constraint='ceil')
                    idx_fin = find_nearest(lbda_mod, lbda_filt[-1], 
                                           constraint='floor')
                    interp_trans = np.interp(lbda_mod[idx_ini:idx_fin], lbda_filt,
                                             trans)
                    num = np.sum(interp_trans*dlbda_mod[idx_ini:idx_fin]*spec_mod[idx_ini:idx_fin])
                    denom = np.sum(interp_trans*dlbda_mod[idx_ini:idx_fin])
                    spec_mod_res[np.where(instru_idx==i)] = num/denom
                else:
                    msg = "instru_fwhm is a {}, while it should be either a"
                    msg+= " scalar or a string"
                    raise TypeError(msg.format(type(instru_fwhm[i-1])))

    return np.array([lbda_obs, spec_mod_res])

    
def interpolate_model(params, grid_param_list, params_em={}, em_grid={}, 
                      em_lines={}, labels=None, model_grid=None, 
                      model_reader=None, interp_order=1, max_dlbda=2e-4,
                      verbose=False):
    """
    Parameters
    ----------
    params : tuple
        Set of models parameters for which the model grid has to be 
        interpolated.
    grid_param_list : list of 1d numpy arrays/lists
        - If list, should contain list/numpy 1d arrays with available grid of 
        model parameters.
        - Note1: model grids should not contain grids on radius and Av, but 
        these should still be passed in initial_state (Av optional).
    params_em : dictionary, opt
        Set of emission line parameters (typically fluxes) for which the model 
        grid has to be interpolated.
    em_grid: dictionary pointing to lists, opt
        Dictionary where each entry corresponds to an emission line and points
        to a list of values to inject for emission line fluxes. For computation 
        efficiency, interpolation will be performed between the points of this 
        grid during the MCMC sampling. Dict entries should match labels and 
        em_lines. Note: length of this dictionary can be different of em_lines;
        i.e. if a line is in em_lines but not in em_grid, it will not be 
        considered an MCMC parameter.
    em_lines: dictionary, opt
        Dictionary of emission lines to be added on top of the model spectrum.
        Each dict entry should be the name of the line, assigned to a tuple of
        3 values: 1) the wavelength (in mu), 2) a string indicating whether line 
        intensity is expressed in flux ('F'), luminosity ('L') or log(L/LSun) 
        ("LogL"), and 3) the latter quantity. The intensity of the emission 
        lines can be sampled by MCMC, in that case the last element of the 
        tuple can be set to None. If not to be sampled, a value for the 
        intensity should be provided (in the same system of units as the model 
        spectra, multiplied by mu). Example:
        em_lines = {'BrG':(2.1667,'F',263)}
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state, that is:
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - next the planet photometric radius 'R', in Jupiter radius
        - (optionally) the flux of emission lines (labels should match those in
        the em_lines dictionary), in units of the model spectrum (times mu)
        - (optionally) the optical extinction 'Av', in mag
        - (optionally) the ratio of total to selective optical extinction 'Rv'
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb
        contribution.
        Note: only necessary if an emission list dictionary is provided.
    model_grid : numpy N-d array
        If provided, should contain the grid of model spectra for each
        free parameter of the given grid. I.e. for a grid of n_T values of Teff 
        and n_g values of Logg, the numpy array should be n_T x n_g x n_ch x 2, 
        where n_ch is the number of wavelengths for the observed spectrum.
        If provided, takes precedence over filename/file_reader which would 
        open and read models at each step of the MCMC.
    model_reader : python routine
        External routine that reads a model file, converts it to required 
        units and returns a 2D numpy array, where the first column corresponds
        to wavelengths, and the second contains model values. Example below.
    interp_order: int, opt, {0,1} 
        0: nearest neighbour model.
        1: Order 1 spline interpolation.
    max_dlbda: float, opt
        Maximum delta lbda in mu allowed if binning of lbda_model is necessary. 
        This is necessary for grids of models (e.g. BT-SETTL) where the wavelength
        sampling is not the same depending on parameters (e.g. between 4000K
        and 4100K models for BT-SETTL): resampling preserving original 
        resolution is too prohibitive computationally.
    verbose: bool, optional
        Whether to print more information during resampling.
        
    Returns
    -------
    model : 2d numpy array
        Interpolated model for input parameters. First column corresponds
        to wavelengths, and the second contains model values.
        
    Example file_reader:
    -------------------
    def _example_file_reader(params):
        '''This is a minimal example for the file_reader routine to be provided 
        as argument to model_interpolation. The routine should only take as 
        inputs grid parameters, and returns as output: both the wavelengths and 
        model values as a 2D numpy array.
        This example assumes the model is in a fits file, that is already a 2D
        numpy array, where the first column is the wavelength, and 2nd column 
        is the corresponding model values.'''
        
        model = open_fits(filename.format(params[0],params[1]))

        return model       
        
    """

    def _example_file_reader(filename):
        """ This is a minimal example for the file_reader routine to be provided 
        as argument to model_interpolation. The routine should take as input a 
        template filename format with blanks and parameters, and return as output 
        the wavelengths and model values as a 2D numpy array.
        This example assumes the model is in a fits file, that is already a 2D
        numpy array, where the first column is the wavelength, and second column 
        is the corresponding model values.
        """
        model = open_fits(filename.format(params[0],params[1]))

        return model

    def _den_to_bin(denary,ndigits=3):
        """
        Convert denary to binary number, keeping n digits for binary (i.e.
        padding with zeros if necessary)
        """
        binary=""  
        while denary>0:  
          #A left shift in binary means /2  
          binary = str(denary%2) + binary  
          denary = denary//2  
        if len(binary) < ndigits:
            pad = '0'*(ndigits-len(binary))
        else:
            pad=''
        return pad+binary  

    n_params = len(grid_param_list)
    n_em = len(em_grid)
    n_params_tot = n_params+n_em

    if interp_order == 0:
        if model_grid is None:
            params_tmp = np.zeros(n_params)    
            for nn in range(n_params):
                params_tmp[nn] = find_nearest(grid_param_list[nn], 
                                              params[nn], output='value')
            lbda, spec = model_reader(params_tmp)
            if n_em>0:
                for ll in range(len(labels)):
                    if labels[ll] in em_grid.keys():
                        key = labels[ll]
                        spec = inject_em_line(em_lines[key][0], params_em[key], 
                                              lbda, spec, em_lines[key][2])
            return lbda, spec
            
        else:
            idx_tmp = []
            counter = 0
            for nn in range(n_params_tot):
                if nn < n_params:
                    idx_tmp.append(find_nearest(grid_param_list[nn], params[nn],
                                                output='index'))
                else:
                    for ll in range(len(labels)):
                        if labels[counter+ll] in em_grid.keys():
                            key = labels[counter+ll]
                            counter+=1
                            break
                    idx_tmp.append(find_nearest(em_grid[key], params_em[key],
                                                output='index'))                    
            idx_tmp = tuple(idx_tmp)
            tmp = model_grid[idx_tmp]
            return tmp[:,0], tmp[:,1]
    
    
    elif abs(interp_order) == 1:
        # first compute new subgrid "coords" for interpolation
        if verbose:
            print("Computing new coords for interpolation")
        constraints = ['floor','ceil']
        new_coords = np.zeros([n_params_tot,1])
        sub_grid_param = np.zeros([n_params_tot,2])
        counter = 0
        for nn in range(n_params_tot):
            if nn < n_params:
                grid_tmp = grid_param_list[nn]
                params_tmp = params[nn]
            else:
                for ll in range(len(labels)):
                    if labels[counter+ll] in em_grid.keys():
                        key = labels[counter+ll]
                        grid_tmp = em_grid[key] 
                        params_tmp = params_em[key]
                        counter+=1
                        break
            for ii in range(2):
                try:
                    sub_grid_param[nn,ii] = find_nearest(grid_tmp, 
                                                         params_tmp,
                                                         constraint=constraints[ii], 
                                                         output='value')
                except:
                    pdb.set_trace()
                
            num = (params_tmp-sub_grid_param[nn,0])
            denom = (sub_grid_param[nn,1]-sub_grid_param[nn,0])
            new_coords[nn,0] = num/denom
            
        if verbose:
            print("Making sub-grid of models")
        sub_grid = []
        sub_grid_lbda = []        
        if model_grid is None:
            ntot_subgrid = 2**n_params_tot
            for dd in range(ntot_subgrid):
                str_indices = _den_to_bin(dd, n_params_tot)
                params_tmp = []
                for nn in range(n_params):
                    params_tmp.append(sub_grid_param[nn,int(str_indices[nn])])
                params_tmp=np.array(params_tmp)
                lbda, spec = model_reader(params_tmp)
                if n_em>0:
                    for nn in range(len(labels)):
                        if labels[nn] in em_grid.keys():
                            key = labels[nn]
                            spec = inject_em_line(em_lines[key][0], 
                                                  params_em[key], lbda, 
                                                  spec, em_lines[key][2])
                sub_grid.append(spec)
                sub_grid_lbda.append(lbda)
            
            # resample to match sparser sampling if required
            nch = np.amin([len(sub_grid_lbda[i]) for i in range(ntot_subgrid)])
            nch_max = np.amax([len(sub_grid_lbda[i]) for i in range(ntot_subgrid)])
            if nch_max != nch:
                min_i = np.argmin([len(sub_grid_lbda[i]) for i in range(ntot_subgrid)])
                min_dlbda = np.amin(sub_grid_lbda[min_i][1:]-sub_grid_lbda[min_i][:-1])
                if min_dlbda < max_dlbda:
                    bin_fac = int(max_dlbda/min_dlbda)
                    if verbose:
                        msg = "Models will be binned in WL by a factor {} to "
                        msg += "min dlbda = {}mu"
                        print(msg.format(bin_fac,max_dlbda))
                    
                    nch = int(len(sub_grid_lbda[min_i])/bin_fac)
                    tmp_spec = []
                    tmp_lbda = []
                    for bb in range(nch):
                        idx_ini = bb*bin_fac
                        idx_fin = (bb+1)*bin_fac
                        tmp_spec.append(np.mean(sub_grid[min_i][idx_ini:idx_fin]))
                        tmp_lbda.append(np.mean(sub_grid_lbda[min_i][idx_ini:idx_fin]))
                    sub_grid[min_i] = np.array(tmp_spec)
                    sub_grid_lbda[min_i] = np.array(tmp_lbda)
                for dd in range(ntot_subgrid):
                    cond = False
                    if len(sub_grid_lbda[dd]) != nch:
                        cond = True
                    else:
                        # np.allclose() or np.array_equal are TOO slow
                        dlbda = sub_grid_lbda[min_i][-1]-sub_grid_lbda[min_i][0]
                        dlbda/=nch
                        if np.sum(sub_grid_lbda[dd]-sub_grid_lbda[min_i])>dlbda:
                            cond = True
                    if cond:
                        if verbose:
                            msg = "Resampling model of different WL sampling. "
                            msg+= "This may take a while for high-res/large WL"
                            msg+= " ranges..."
                            print(msg)
                        res = resample_model(sub_grid_lbda[min_i], 
                                             sub_grid_lbda[dd], 
                                             sub_grid[dd],
                                             no_constraint=True,
                                             verbose=verbose)
                        sub_grid_lbda[dd], sub_grid[dd] = res

            # Create array with dimensions 'dims' for each wavelength
            final_dims = tuple([nch]+[2]*n_params_tot)
            sub_grid = np.array(sub_grid)
            sub_grid_lbda = np.array(sub_grid_lbda)
            sub_grid = np.swapaxes(sub_grid,0,1)            
            sub_grid_lbda = np.swapaxes(sub_grid_lbda,0,1)
            sub_grid = sub_grid.reshape(final_dims)
            sub_grid_lbda = sub_grid_lbda.reshape(final_dims)
        
        else:
            constraints = ['floor','ceil']
            sub_grid_idx = np.zeros([n_params_tot,2], dtype=np.int32)
            #list_idx = []
            counter = 0
            for nn in range(n_params_tot):
                if nn < n_params:
                    grid_tmp = grid_param_list[nn]
                    params_tmp = params[nn]
                else:
                    for ll in range(len(labels)):
                        if labels[counter+ll] in em_grid.keys():
                            key = labels[counter+ll]
                            grid_tmp = em_grid[key] 
                            params_tmp = params_em[key]
                            counter+=1
                            break
                for ii in range(2):
                    sub_grid_idx[nn,ii]=find_nearest(grid_tmp, params_tmp,
                                                     constraint=constraints[ii], 
                                                     output='index')
            for dd in range(2**n_params_tot):
                str_indices = _den_to_bin(dd, n_params_tot)
                idx_tmp = []
                for nn in range(n_params_tot):
                    idx_tmp.append(sub_grid_idx[nn,int(str_indices[nn])])
                    #idx_tmp = sub_grid_idx[nn,int(str_indices[nn])]
                #list_idx.append(idx_tmp)
                #list_idx=np.array(list_idx)
                sub_grid.append(model_grid[tuple(idx_tmp)])
            
            # first reshape
            sub_grid = np.array(sub_grid)
            dims = tuple([2]*n_params_tot+[sub_grid.shape[-2]]+[sub_grid.shape[-1]])
            sub_grid = sub_grid.reshape(dims)
            sub_grid = np.moveaxis(sub_grid,-1,0) # make last dim (lbda vs flux) come first
            sub_grid_lbda = sub_grid[0]
            sub_grid = sub_grid[1]
            # move again axis to have nch as first axis
            sub_grid = np.moveaxis(sub_grid,-1,0)
            sub_grid_lbda = np.moveaxis(sub_grid_lbda,-1,0)
            nch = sub_grid.shape[0]
            
        interp_model = np.zeros(nch)
        interp_lbdas = np.zeros(nch)
        
        if interp_order == -1:
            sub_grid = np.log10(sub_grid)
        
        for cc in range(nch):
            interp_model[cc] = map_coordinates(sub_grid[cc], new_coords, 
                                               order=abs(interp_order))
            interp_lbdas[cc] = map_coordinates(sub_grid_lbda[cc], new_coords, 
                                               order=abs(interp_order))
            
        if interp_order == -1:
            interp_model = np.power(10,interp_model)
            
        return interp_lbdas, interp_model
    
    else:
        msg = "Interpolation order not allowed. Only -1, 0 or 1 accepted"
        raise TypeError(msg)


"""
ROC curves generation.
"""
__all__ = ['EvalRoc',
           'compute_binary_map']

import copy
import numpy as np
import matplotlib.pyplot as plt
from hciplot import plot_frames
from scipy import stats
from photutils import detect_sources
from munch import Munch
from ..pca.svd import SVDecomposer
from ..var import frame_center, get_annulus_segments
from ..conf import time_ini, timing, Progressbar
from ..var import get_circle
from .fakecomp import cube_inject_companions


# TODO: remove the munch dependency
class EvalRoc(object):
    """
    Class for the generation of receiver operating characteristic (ROC) curves.
    """

    COLOR_1 = "#d62728"  # CADI
    COLOR_2 = "#ff7f0e"  # PCA
    COLOR_3 = "#2ca02c"  # LLSG
    COLOR_4 = "#9467bd"  # SODIRF
    COLOR_5 = "#1f77b4"  # SODINN
    SYMBOL_1 = "^"  # CADI
    SYMBOL_2 = "X"  # PCA
    SYMBOL_3 = "P"  # LLSG
    SYMBOL_4 = "s"  # SODIRF
    SYMBOL_5 = "p"  # SODINN
    # For model PSF subtraction algos that rely on a S/N map
    THRESHOLDS_05_5 = [0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5]
    # For algos that output a likelihood or probability map
    THRESHOLDS_01_099 = np.linspace(0.1, 0.99, 10).tolist()

    def __init__(self, dataset, plsc=0.0272, n_injections=100, inrad=8,
                 outrad=12, dist_flux=("uniform", 2, 500), mask=None):
        """
        [...]
        dist_flux : tuple ('method', *args)
            'method' can be a string, e.g:
                ("skewnormal", skew, mean, var)
                ("uniform", low, high)
                ("normal", loc, scale)
            or a function.
        [...]
        """
        self.dataset = dataset
        self.plsc = plsc
        self.n_injections = n_injections
        self.inrad = inrad
        self.outrad = outrad
        self.dist_flux = dist_flux
        self.mask = mask
        self.methods = []

    def add_algo(self, name, algo, color, symbol, thresholds):
        """
        Parameters
        ----------
        algo : HciPostProcAlgo
        thresholds : list of lists

        """
        self.methods.append(Munch(algo=algo, name=name, color=color,
                                  symbol=symbol, thresholds=thresholds))

    def inject_and_postprocess(self, patch_size, cevr=0.9,
                               expvar_mode='annular', nproc=1):
        """

        Notes
        -----
        # TODO `methods` are not returned inside `results` and are *not* saved!
        # TODO order of parameters for `skewnormal` `dist_flux` changed! (was [3], [1], [2])
        # TODO `save` not implemented

        """
        from .. import hci_postproc

        starttime = time_ini()

        frsize = self.dataset.cube.shape[1]

        # ===== number of PCs for PCA / rank for LLSG
        if cevr is not None:
            svdecomp = SVDecomposer(self.dataset.cube, mode=expvar_mode,
                                    inrad=self.inrad, outrad=self.outrad,
                                    svd_mode='lapack', verbose=False)
            _ = svdecomp.get_cevr(ncomp_list=None)
            ratio_cumsum = svdecomp.cevr
            self.optpcs = np.searchsorted(ratio_cumsum, cevr) + 1
            print("{}% of CEVR with {} PCs".format(cevr, self.optpcs))

            # for m in methods:
            #     if hasattr(m, "ncomp") and m.ncomp is None:  # PCA
            #         m.ncomp = self.optpcs
            #
            #     if hasattr(m, "rank") and m.rank is None:  # LLSG
            #         m.rank = self.optpcs

            #
            #   ------> this should be moved inside the HCIPostProcAlgo classes!
            #
        # Getting indices in annulus
        width = self.outrad - self.inrad
        yy, xx = get_annulus_segments(self.dataset.cube[0], self.inrad,
                                      width)[0]
        num_patches = yy.shape[0]

        # Defining Fluxes according to chosen distribution
        dist_fkt = dict(skewnormal=stats.skewnorm.rvs,
                        normal=np.random.normal,
                        uniform=np.random.uniform).get(self.dist_flux[0],
                                                       self.dist_flux[0])

        self.fluxes = dist_fkt(*self.dist_flux[1:], size=self.n_injections)
        self.fluxes.sort()
        inds_inj = np.random.randint(0, num_patches, size=self.n_injections)

        self.dists = []
        self.thetas = []
        for m in range(self.n_injections):
            injx = xx[inds_inj[m]]
            injy = yy[inds_inj[m]]
            injx -= frame_center(self.dataset.cube[0])[1]
            injy -= frame_center(self.dataset.cube[0])[0]
            dist = np.sqrt(injx**2 + injy**2)
            theta = np.mod(np.arctan2(injy, injx) / np.pi * 180, 360)
            self.dists.append(dist)
            self.thetas.append(theta)

        for m in self.methods:
            m.frames = []
            m.probmaps = []

        self.list_xy = []

        # Injections
        for n in Progressbar(range(self.n_injections), desc="injecting"):
            cufc, cox, coy = _create_synt_cube(self.dataset.cube,
                                               self.dataset.psf,
                                               self.dataset.angles, self.plsc,
                                               theta=self.thetas[n],
                                               flux=self.fluxes[n],
                                               dist=self.dists[n],
                                               verbose=False)
            cox = int(np.round(cox))
            coy = int(np.round(coy))
            self.list_xy.append((cox, coy))

            for m in self.methods:
                # TODO: this is not elegant at all.
                # shallow copy. Should not copy e.g. the cube in memory,
                # just reference it.
                algo = copy.copy(m.algo)
                _dataset = copy.copy(self.dataset)
                _dataset.cube = cufc

                if isinstance(algo, hci_postproc.HCIPca):
                    algo.ncomp = self.optpcs
                # elif isinstance(algo, hci_postproc.HCILLSG):
                #     algo.rank = self.optpcs

                algo.run(dataset=_dataset, verbose=False)
                algo.make_snrmap(approximated=True, nproc=nproc, verbose=False)

                m.frames.append(algo.frame_final)
                m.probmaps.append(algo.snr_map)

        timing(starttime)

    def compute_tpr_fps(self, **kwargs):
        """
        Calculate number of dets/fps for every injection/method/threshold.

        Take the probability maps and the desired thresholds for every method,
        and calculates the binary map, number of detections and FPS using
        ``compute_binary_map``. Sets each methods ``detections``, ``fps`` and
        ``bmaps`` attributes.

        Parameters
        ----------
        **kwargs : keyword arguments
            Passed to ``compute_binary_map``

        """
        starttime = time_ini()

        for m in self.methods:
            m.detections = []
            m.fps = []
            m.bmaps = []

        print('Evaluating injections:')
        for i in Progressbar(range(self.n_injections)):
            x, y = self.list_xy[i]

            for m in self.methods:
                dets, fps, bmaps = compute_binary_map(
                    m.probmaps[i], m.thresholds, fwhm=self.dataset.fwhm,
                    injections=(x, y), **kwargs
                )
                m.detections.append(dets)
                m.fps.append(fps)
                m.bmaps.append(bmaps)

        timing(starttime)

    def plot_detmaps(self, i=None, thr=9, dpi=100,
                     axis=True, grid=False, vmin=-10, vmax='max',
                     plot_type="horiz"):
        """
        Plot the detection maps for one injection.

        Parameters
        ----------
        i : int or None, optional
            Index of the injection, between 0 and self.n_injections. If None,
            takes the 30st injection, or if there are less injections, the
            middle one.
        thr : int, optional
            Index of the threshold.
        dpi, axis, grid, vmin, vmax
            Passed to ``pp_subplots``
        plot_type : {"horiz" or "vert"}, optional
            Plot type.

            ``horiz``
                One row per algorithm (frame, probmap, binmap)
            ``vert``
                1 row for final frames, 1 row for probmaps and 1 row for binmaps

        """
        # input parameters
        if i is None:
            if len(self.list_xy) > 30:
                i = 30
            else:
                i = len(self.list_xy) // 2

        if vmax == 'max':
            # TODO: document this feature.
            vmax = np.concatenate([m.frames[i] for m in self.methods if
                                   hasattr(m, "frames") and
                                   len(m.frames) >= i]).max()/2

        # print information
        print('X,Y: {}'.format(self.list_xy[i]))
        print('dist: {:.3f}, flux: {:.3f}'.format(self.dists[i],
                                                  self.fluxes[i]))
        print()

        if plot_type in [1, "horiz"]:
            for m in self.methods:
                print('detection state: {} | false postives: {}'.format(
                    m.detections[i][thr], m.fps[i][thr]))
                labels = ('{} frame'.format(m.name), '{} S/Nmap'.format(m.name),
                          'Thresholded at {:.1f}'.format(m.thresholds[thr]))
                plot_frames((m.frames[i] if len(m.frames) >= i else
                            np.zeros((2, 2)), m.probmaps[i], m.bmaps[i][thr]),
                            label=labels, dpi=dpi, horsp=0.2, axis=axis,
                            grid=grid, cmap=['viridis', 'viridis', 'gray'])

        elif plot_type in [2, "vert"]:
            labels = tuple('{} frame'.format(m.name) for m in self.methods if
                           hasattr(m, "frames") and len(m.frames) >= i)
            plot_frames(tuple(m.frames[i] for m in self.methods if
                        hasattr(m, "frames") and len(m.frames) >= i),
                        dpi=dpi, label=labels, vmax=vmax, vmin=vmin, axis=axis,
                        grid=grid)

            plot_frames(tuple(m.probmaps[i] for m in self.methods), dpi=dpi,
                        label=tuple(['{} S/Nmap'.format(m.name) for m in
                                     self.methods]), axis=axis, grid=grid)

            for m in self.methods:
                msg = '{} detection: {}, FPs: {}'
                print(msg.format(m.name, m.detections[i][thr], m.fps[i][thr]))

            labels = tuple('Thresholded at {:.1f}'.format(m.thresholds[thr])
                           for m in self.methods)
            plot_frames(tuple(m.bmaps[i][thr] for m in self.methods),
                        dpi=dpi, label=labels, axis=axis, grid=grid,
                        colorbar=False, cmap='bone')
        else:
            raise ValueError("`plot_type` unknown")

    def plot_roc_curves(self, dpi=100, figsize=(5, 5), xmin=None, xmax=None,
                        ymin=-0.05, ymax=1.02, xlog=True, label_skip_one=False,
                        legend_loc='lower right', legend_size=6,
                        show_data_labels=True, hide_overlap_label=True,
                        label_gap=(0, -0.028), save_plot=False, label_params={},
                        line_params={}, marker_params={}, verbose=True):
        """
        Parameters
        ----------


        Returns
        -------
        None, but modifies `methods`: adds .tpr and .mean_fps attributes

        Notes
        -----
        # TODO: load `roc_injections` and `roc_tprfps` from file (`load_res`)
        # TODO: print flux distro information (is it actually stored in inj?
        What to do with functions, do they pickle?)
        # TODO: hardcoded `methodconf`?

        """
        labelskw = dict(alpha=1, fontsize=5.5, weight="bold", rotation=0,
                        annotation_clip=True)
        linekw = dict(alpha=0.2)
        markerkw = dict(alpha=0.5, ms=3)
        labelskw.update(label_params)
        linekw.update(line_params)
        markerkw.update(marker_params)
        n_thresholds = len(self.methods[0].thresholds)

        if verbose:
            print('{} injections'.format(self.n_injections))
            # print('Flux distro : {} [{}:{}]'.format(roc_injections.flux_distribution,
            #                                         roc_injections.fluxp1, roc_injections.fluxp2))
            print('Annulus from {} to {} pixels'.format(self.inrad,
                                                        self.outrad))

        fig = plt.figure(figsize=figsize, dpi=dpi)
        ax = fig.add_subplot(111)

        if not isinstance(label_skip_one, (list, tuple)):
            label_skip_one = [label_skip_one]*len(self.methods)
        labels = []

        # methodconf = {"CADI": dict(color="#d62728", symbol="^"),
        #              "PCA": dict(color="#ff7f0e", symbol="X"),
        #              "LLSG": dict(color="#2ca02c", symbol="P"),
        #              "SODIRF": dict(color="#9467bd", symbol="s"),
        #              "SODINN": dict(color="#1f77b4", symbol="p"),
        #              "SODINN-pw": dict(color="#1f77b4", symbol="p")
        #             }  # maps m.name to plot style

        for i, m in enumerate(self.methods):

            if not hasattr(m, "detections") or not hasattr(m, "fps"):
                raise AttributeError("method #{} has no detections/fps. Run"
                                     "`compute_tpr_fps` first.".format(i))

            m.tpr = np.zeros((n_thresholds))
            m.mean_fps = np.zeros((n_thresholds))

            for j in range(n_thresholds):
                m.tpr[j] = np.asarray(m.detections)[:, j].tolist().count(1) / \
                           self.n_injections
                m.mean_fps[j] = np.asarray(m.fps)[:, j].mean()

            plt.plot(m.mean_fps, m.tpr, '--', color=m.color, **linekw)
            plt.plot(m.mean_fps, m.tpr, m.symbol, label=m.name, color=m.color,
                     **markerkw)

            if show_data_labels:
                if label_skip_one[i]:
                    lab_x = m.mean_fps[1::2]
                    lab_y = m.tpr[1::2]
                    thr = m.thresholds[1::2]
                else:
                    lab_x = m.mean_fps
                    lab_y = m.tpr
                    thr = m.thresholds

                for i, xy in enumerate(zip(lab_x + label_gap[0],
                                           lab_y + label_gap[1])):
                    labels.append(ax.annotate('{:.2f}'.format(thr[i]),
                                  xy=xy, xycoords='data', color=m.color,
                                              **labelskw))
                    # TODO: reverse order of `self.methods` for better annot. z-index?

        plt.legend(loc=legend_loc, prop={'size': legend_size})
        if xlog:
            ax.set_xscale("symlog")
        plt.ylim(ymin=ymin, ymax=ymax)
        plt.xlim(xmin=xmin, xmax=xmax)
        plt.ylabel('TPR')
        plt.xlabel('Full-frame mean FPs')
        plt.grid(alpha=0.4)

        if show_data_labels:
            mask = np.zeros(fig.canvas.get_width_height(), bool)

            fig.canvas.draw()

            for label in labels:
                bbox = label.get_window_extent()
                negpad = -2
                x0 = int(bbox.x0) + negpad
                x1 = int(np.ceil(bbox.x1)) + negpad
                y0 = int(bbox.y0) + negpad
                y1 = int(np.ceil(bbox.y1)) + negpad

                s = np.s_[x0:x1, y0:y1]
                if np.any(mask[s]):
                    if hide_overlap_label:
                        label.set_visible(False)
                else:
                    mask[s] = True

        if save_plot:
            if isinstance(save_plot, str):
                plt.savefig(save_plot, dpi=dpi, bbox_inches='tight')
            else:
                plt.savefig('roc_curve.pdf', dpi=dpi, bbox_inches='tight')


def compute_binary_map(frame, thresholds, injections, fwhm, npix=1,
                       overlap_threshold=0.7, max_blob_fact=2, plot=False,
                       debug=False):
    """
    Take a list of ``thresholds``, create binary maps and counts detections/fps.
    A blob which is "too big" is split into apertures, and every aperture adds
    one 'false positive'.

    Parameters
    ----------
    frame : numpy ndarray
        Detection map.
    thresholds : list or numpy ndarray
        List of thresholds (detection criteria).
    injections : tuple, list of tuples
        Coordinates (x,y) of the injected companions. Also accepts 1d/2d
        ndarrays.
    fwhm : float
        FWHM, used for obtaining the size of the circular aperture centered at
        the injection position (and measuring the overlapping with found blobs).
        The circular aperture has 2 * FWHM in diameter.
    npix : int, optional
        The number of connected pixels, each greater than the given threshold,
        that an object must have to be detected. ``npix`` must be a positive
        integer. Passed to ``detect_sources`` function from ``photutils``.
    overlap_threshold : float
        Percentage of overlap a blob has to have with the aperture around an
        injection.
    max_blob_fact : float
        Maximum size of a blob (in multiples of the resolution element) before
        it is considered as "too big" (= non-detection).
    plot : bool, optional
        If True, a final resulting plot summarizing the results will be shown.
    debug : bool, optional
        For showing optional information.

    Returns
    -------
    list_detections : list of int
        List of detection count for each threshold.
    list_fps : list of int
        List of false positives count for each threshold.
    list_binmaps : list of 2d ndarray
        List of binary maps: detection maps thresholded for each threshold
        value.

    """
    def _overlap_injection_blob(injection, fwhm, blob_mask):
        """
        Parameters
        ----------
        injection: tuple (y,x)
        fwhm : float
        blob_mask : 2d bool ndarray

        Returns
        -------
        overlap_fact : float between 0 and 1
            Percentage of the area overlap. If the blob is smaller than the
            resolution element, this is ``intersection_area / blob_area``,
            otherwise ``intersection_area / resolution_element``.

        """
        if len(injections[0]) > 0:
            injection_mask = get_circle(np.ones_like(blob_mask), radius=fwhm,
                                        cy=injection[1], cx=injection[0],
                                        mode="mask")
        else:
            injection_mask = np.zeros_like(blob_mask)
        intersection = injection_mask & blob_mask
        smallest_area = min(blob_mask.sum(), injection_mask.sum())
        return intersection.sum() / smallest_area

    # --------------------------------------------------------------------------
    list_detections = []
    list_fps = []
    list_binmaps = []
    sizey, sizex = frame.shape
    cy, cx = frame_center(frame)
    reselem_mask = get_circle(frame, radius=fwhm, cy=cy, cx=cx, mode="val")
    npix_circ_aperture = reselem_mask.shape[0]

    # normalize injections: accepts combinations of 1d/2d and tuple/list/array.
    injections = np.asarray(injections)
    if injections.ndim == 1:
        injections = np.array([injections])

    for ithr, threshold in enumerate(thresholds):
        if debug:
            print("\nprocessing threshold #{}: {}".format(ithr + 1, threshold))

        segments = detect_sources(frame, threshold, npix, connectivity=4)
        binmap = (segments.data != 0)

        if debug:
            plot_frames((segments.data, binmap), cmap=('tab20b', 'binary'),
                        circle=tuple(tuple(xy) for xy in injections),
                        circle_radius=fwhm, circle_alpha=0.6,
                        label=("segmentation map", "binary map"))

        detections = 0
        fps = 0

        for segment in segments.segments:
            label = segment.label
            blob_mask = (segments.data == label)
            blob_area = segment.area

            if debug:
                lab = "blob #{}, area={}px**2".format(label, blob_area)
                plot_frames(blob_mask, circle_radius=fwhm, circle_alpha=0.6,
                            circle=tuple(tuple(xy) for xy in injections),
                            cmap='binary', label_size=8, label=lab,
                            size_factor=3)

            for iinj, injection in enumerate(injections):
                if len(injections[0]) > 0:  # checking injections is not empty
                    if injection[0] > sizex or injection[1] > sizey:
                        raise ValueError("Wrong coordinates in `injections`")

                    if debug:
                        print("\ttesting injection #{} at {}".format(iinj + 1,
                                                                     injection))

                if blob_area > max_blob_fact * npix_circ_aperture:
                    number_of_apertures_in_blob = blob_area / npix_circ_aperture
                    fps += number_of_apertures_in_blob  # float, rounded at end
                    if debug:
                        print("\tblob is too big (+{:.0f} fps)"
                              "".format(number_of_apertures_in_blob))
                        print("\tskipping all other injections")
                    # continue with next blob, do not check other injections
                    break

                overlap = _overlap_injection_blob(injection, fwhm, blob_mask)
                if overlap > overlap_threshold:
                    if debug:
                        print("\toverlap of {}! (+1 detection)"
                              "".format(overlap))

                    detections += 1
                    # continue with next blob, do not check other injections
                    break

                if debug:
                    print("\toverlap of {} -> do nothing".format(overlap))

            else:
                if debug:
                    print("\tdid not find a matching injection for this "
                          "blob (+1 fps)")
                fps += 1

        if debug:
            print("done with threshold #{}".format(ithr))
            print("result: {} detections, {} fps".format(detections, fps))

        fps = np.round(fps).astype(int).item()  # -> python `int`

        list_detections.append(detections)
        list_binmaps.append(binmap)
        list_fps.append(fps)

    if plot:
        labs = tuple(str(det) + ' detections' + '\n' + str(fps) +
                     ' false positives' for det, fps in zip(list_detections,
                                                            list_fps))
        if len(injections[0]) > 0:
            circles = tuple(tuple(xy) for xy in injections)
        else:
            circles = None
        plot_frames(tuple(list_binmaps), title='Final binary maps', label=labs,
                    label_size=8, cmap='binary', circle_alpha=0.8,
                    circle=circles, circle_radius=fwhm,
                    circle_color='deepskyblue', axis=False)

    return list_detections, list_fps, list_binmaps


def _create_synt_cube(cube, psf, ang, plsc, dist, flux, theta=None,
                      verbose=False):
    """
    """
    centy_fr, centx_fr = frame_center(cube[0])
    if theta is None:
        np.random.seed()
        theta = np.random.randint(0, 360)

    posy = dist * np.sin(np.deg2rad(theta)) + centy_fr
    posx = dist * np.cos(np.deg2rad(theta)) + centx_fr
    if verbose:
        print('Theta:', theta)
        print('Flux_inj:', flux)
    cubefc = cube_inject_companions(cube, psf, ang, flevel=flux, plsc=plsc,
                                    rad_dists=[dist], n_branches=1, theta=theta,
                                    verbose=verbose)
    return cubefc, posx, posy
